# coding: utf-8
"""

"""

import datetime
import json
import os
import typing

import itsdangerous
import flask
import flask_login
import wtforms
import werkzeug
from flask_babel import _
from flask_wtf import FlaskForm
import pytz
from wtforms import StringField, SelectMultipleField, BooleanField, MultipleFileField, IntegerField, SelectField
from wtforms.validators import DataRequired, ValidationError, InputRequired

from . import frontend
from .objects.permissions import get_object_if_current_user_has_read_permissions, get_fed_object_if_current_user_has_read_permissions
from .objects.view import get_project_if_it_exists
from .utils import get_user_if_exists
from ..logic.action_permissions import get_user_action_permissions
from ..logic.components import get_component
from ..logic.instruments import get_instrument, create_instrument, update_instrument, set_instrument_responsible_users, get_instruments, set_instrument_location, get_instrument_object_links, set_instrument_object
from ..logic.instrument_log_entries import get_instrument_log_entries, create_instrument_log_entry, get_instrument_log_file_attachment, create_instrument_log_file_attachment, create_instrument_log_object_attachment, get_instrument_log_categories, create_instrument_log_category, update_instrument_log_category, delete_instrument_log_category, update_instrument_log_entry, hide_instrument_log_file_attachment, hide_instrument_log_object_attachment, get_instrument_log_entry, get_instrument_log_object_attachment
from ..logic.instrument_translations import get_instrument_translations_for_instrument, set_instrument_translation, delete_instrument_translation
from ..logic.languages import get_languages, get_language, Language, get_user_language
from ..logic.actions import get_actions, get_action
from ..logic.action_types import ActionType, get_action_types, get_action_type
from ..logic.errors import InstrumentDoesNotExistError, InstrumentLogFileAttachmentDoesNotExistError, ObjectDoesNotExistError, UserDoesNotExistError, InstrumentLogEntryDoesNotExistError, InstrumentLogObjectAttachmentDoesNotExistError, InstrumentObjectLinkAlreadyExistsError
from ..logic.favorites import get_user_favorite_instrument_ids, get_user_favorite_action_ids
from ..logic.markdown_images import mark_referenced_markdown_images_as_permanent
from ..logic.users import get_users, get_user
from ..logic.objects import get_object
from ..logic.object_permissions import get_object_info_with_permissions, get_user_object_permissions
from ..logic.settings import get_user_settings, set_user_settings
from ..logic.shares import get_shares_for_object
from ..logic.locations import get_location, get_object_location_assignment
from .users.forms import ToggleFavoriteInstrumentForm
from .utils import check_current_user_is_not_readonly, generate_qrcode, get_locations_form_data, parse_filter_id_params, build_modified_url
from ..utils import FlaskResponseT
from ..logic.utils import get_translated_text
from ..logic.markdown_to_html import markdown_to_safe_html
from ..logic.topics import set_instrument_topics, get_topics, get_topic
from .validators import MultipleObjectIdValidator

from ..models.permissions import Permissions
from ..models.instrument_log_entries import InstrumentLogCategoryTheme

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

INSTRUMENT_LIST_FILTER_PARAMETERS = (
    'instrument_list_filters',
    'topic_ids'
)


class MultipleIntegerField(SelectMultipleField):  # type: ignore[misc]
    def pre_validate(self, form: wtforms.Form) -> None:
        if self.data:
            for d in self.data:
                try:
                    int(d)
                except ValueError:
                    raise ValidationError("Invalid value")


class InstrumentLogEntryForm(FlaskForm):
    content = StringField()
    content_is_markdown = BooleanField()
    files = MultipleFileField()
    objects = SelectMultipleField(validators=[MultipleObjectIdValidator(Permissions.READ)], validate_choice=False)
    categories = SelectMultipleField()
    log_entry_id = IntegerField()
    existing_files = MultipleIntegerField()
    existing_objects = MultipleIntegerField()
    has_event_utc_datetime = BooleanField()
    event_utc_datetime = StringField()

    def validate_event_utc_datetime(form, field: wtforms.BooleanField) -> None:
        if field.data:
            try:
                language = get_user_language(flask_login.current_user)
                parsed_datetime = datetime.datetime.strptime(field.data, language.datetime_format_datetime)
                # convert datetime to utc
                local_datetime = pytz.timezone(flask_login.current_user.timezone or 'UTC').localize(parsed_datetime)
                utc_datetime = local_datetime.astimezone(pytz.utc)
                if abs(utc_datetime.year - datetime.date.today().year) > 1000:
                    raise ValueError()
                field.data = utc_datetime.strftime('%Y-%m-%d %H:%M:%S')
            except Exception:
                raise ValidationError("Invalid datetime")
        else:
            field.data = None


class InstrumentLogOrderForm(FlaskForm):
    ascending = BooleanField()
    attribute = StringField(validators=[DataRequired()])

    def validate_attribute(form, field: wtforms.StringField) -> None:
        if field.data not in ('datetime', 'username'):
            raise ValidationError("invalid log order attribute")


class ObjectLinkForm(FlaskForm):
    object_id = IntegerField(validators=[InputRequired()])


def _parse_instrument_list_filters(
        params: werkzeug.datastructures.MultiDict[str, str],
        valid_topic_ids: typing.List[int]
) -> typing.Tuple[
    bool,
    typing.Optional[typing.List[int]],
]:
    FALLBACK_RESULT = False, None
    success, filter_topic_ids = parse_filter_id_params(
        params=params,
        param_aliases=['topic_ids'],
        valid_ids=valid_topic_ids,
        id_map={},
        multi_params_error='',
        parse_error=_('Unable to parse topic IDs.'),
        invalid_id_error=_('Invalid topic ID.')
    )
    if not success:
        return FALLBACK_RESULT

    return (
        True,
        filter_topic_ids,
    )


@frontend.route('/instruments/')
@flask_login.login_required
def instruments() -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    all_instruments = get_instruments()
    instruments = []
    user_has_admin_permissions = flask_login.current_user.has_admin_permissions
    for instrument in all_instruments:
        if instrument.is_hidden and not user_has_admin_permissions and flask_login.current_user not in instrument.responsible_users:
            continue
        instruments.append(instrument)
    user_favorite_instrument_ids = get_user_favorite_instrument_ids(flask_login.current_user.id)

    topics = get_topics()
    valid_topic_ids = [topic.id for topic in topics]

    if 'instrument_list_filters' in flask.request.args or any(any(flask.request.args.getlist(param)) for param in INSTRUMENT_LIST_FILTER_PARAMETERS):
        (
            success,
            filter_topic_ids,
        ) = _parse_instrument_list_filters(
            params=flask.request.args,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
    else:
        filter_topic_ids = get_user_settings(user_id=flask_login.current_user.id)['DEFAULT_INSTRUMENT_LIST_FILTERS'].get('filter_topic_ids', [])

    if filter_topic_ids is None or flask.current_app.config['DISABLE_TOPICS']:
        filter_topic_ids = []

    if filter_topic_ids:
        instruments = [
            instrument
            for instrument in instruments
            if any(topic in [topic.id for topic in instrument.topics] for topic in filter_topic_ids)
        ]

    filter_topic_infos = []
    if filter_topic_ids:
        for topic_id in filter_topic_ids:
            topic = get_topic(topic_id)
            filter_topic_infos.append({
                'name': get_translated_text(topic.name, default=_('Unnamed Topic')),
                'url': flask.url_for('.topic', topic_id=topic_id)
            })

    # Sort by: favorite / not favorite, instrument name
    instruments.sort(key=lambda instrument: (
        0 if instrument.id in user_favorite_instrument_ids else 1,
        get_translated_text(instrument.name).lower()
    ))
    toggle_favorite_instrument_form = ToggleFavoriteInstrumentForm()
    return flask.render_template(
        'instruments/instruments.html',
        instruments=instruments,
        user_favorite_instrument_ids=user_favorite_instrument_ids,
        toggle_favorite_instrument_form=toggle_favorite_instrument_form,
        get_user=get_user,
        get_component=get_component,
        topics=topics,
        filter_topic_infos=filter_topic_infos,
        filter_topic_ids=filter_topic_ids,
    )


@frontend.route('/instruments/', methods=['POST'])
@flask_login.login_required
def save_instrument_list_defaults() -> FlaskResponseT:
    if 'save_default_instrument_filters' in flask.request.form:
        topics = get_topics()
        valid_topic_ids = [topic.id for topic in topics]
        success, filter_topic_ids = _parse_instrument_list_filters(
            params=flask.request.form,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_INSTRUMENT_LIST_FILTERS': {
                    'filter_topic_ids': filter_topic_ids,
                }
            }
        )
        return flask.redirect(build_modified_url('.instruments', blocked_parameters=INSTRUMENT_LIST_FILTER_PARAMETERS))
    if 'clear_default_instrument_filters' in flask.request.form:
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_INSTRUMENT_LIST_FILTERS': {}
            }
        )
        return flask.redirect(build_modified_url('.instruments', blocked_parameters=INSTRUMENT_LIST_FILTER_PARAMETERS))
    return flask.abort(400)


@frontend.route('/instruments/<int:instrument_id>', methods=['GET', 'POST'])
@flask_login.login_required
def instrument(instrument_id: int) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    user_language = get_user_language(flask_login.current_user)

    is_instrument_responsible_user = any(
        responsible_user.id == flask_login.current_user.id
        for responsible_user in instrument.responsible_users
    )
    already_linked_object_ids = []
    linkable_action_ids = []
    if is_instrument_responsible_user and not flask_login.current_user.is_readonly:
        object_link_form = ObjectLinkForm()
        if instrument.object_id is None:
            already_linked_object_ids = [link[1] for link in get_instrument_object_links()]
            for action_type in get_action_types():
                if action_type.enable_instrument_link:
                    linkable_action_ids.extend([
                        action.id
                        for action in get_actions(action_type_id=action_type.id)
                    ])
    else:
        object_link_form = None
    user_favorite_action_ids = get_user_favorite_action_ids(flask_login.current_user.id)

    template_kwargs = {
        "get_user": get_user_if_exists,
        "get_component": get_component,
        "user_favorite_action_ids": user_favorite_action_ids,
    }
    linked_object = None
    show_object_title = False
    if instrument.object_id is not None:
        linked_object_permissions = get_user_object_permissions(object_id=instrument.object_id, user_id=flask_login.current_user.id)
        if Permissions.READ in linked_object_permissions:
            linked_object = get_object(instrument.object_id)
            if linked_object.data and linked_object.schema:
                show_object_title = get_user_settings(flask_login.current_user.id)['SHOW_OBJECT_TITLE']
                # various getters used by object view
                template_kwargs.update({
                    "get_object": get_object,
                    "get_object_if_current_user_has_read_permissions": get_object_if_current_user_has_read_permissions,
                    "get_fed_object_if_current_user_has_read_permissions": get_fed_object_if_current_user_has_read_permissions,
                    "get_location": get_location,
                    "get_object_location_assignment": get_object_location_assignment,
                    "get_project": get_project_if_it_exists,
                    "get_action_type": get_action_type,
                    "get_shares_for_object": get_shares_for_object,
                })

    if is_instrument_responsible_user or instrument.users_can_view_log_entries:
        instrument_log_entries = get_instrument_log_entries(instrument_id)
        attached_object_ids = set()
        for log_entry in instrument_log_entries:
            for object_attachment in log_entry.object_attachments:
                attached_object_ids.add(object_attachment.object_id)
        attached_object_infos = get_object_info_with_permissions(
            user_id=flask_login.current_user.id,
            permissions=Permissions.READ,
            object_ids=list(attached_object_ids)
        )
        attached_object_names = {
            object_info.object_id: get_translated_text(object_info.name_json)
            for object_info in attached_object_infos
        }
        instrument_log_user_ids = {
            log_entry.user_id
            for log_entry in instrument_log_entries
        }
        instrument_log_users = [
            get_user(user_id)
            for user_id in instrument_log_user_ids
        ]
        instrument_log_users.sort(key=lambda user: (user.name, user.id))
    else:
        instrument_log_entries = None
        instrument_log_users = []
        attached_object_names = {}
    if is_instrument_responsible_user or instrument.users_can_create_log_entries:
        serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
        token = serializer.dumps([flask_login.current_user.id, instrument_id])
        mobile_upload_url = flask.url_for('.instrument_log_mobile_file_upload', instrument_id=instrument_id, token=token, _external=True)
        mobile_upload_qrcode = generate_qrcode(mobile_upload_url, should_cache=False)
    else:
        mobile_upload_url = None
        mobile_upload_qrcode = None
    instrument_log_entry_form = InstrumentLogEntryForm()
    instrument_log_entry_form.objects.choices = []
    instrument_log_categories = get_instrument_log_categories(instrument_id)
    instrument_log_entry_form.categories.choices = [
        (str(category.id), category.title)
        for category in instrument_log_categories
    ]
    instrument_log_entry_form.event_utc_datetime.format = user_language.datetime_format_datetime
    instrument_log_entry_form.event_utc_datetime.format_moment = user_language.datetime_format_moment
    if instrument_log_entry_form.validate_on_submit():
        check_current_user_is_not_readonly()
        if 'action_edit_content' in flask.request.form:
            if is_instrument_responsible_user or instrument.users_can_view_log_entries:
                try:
                    log_entry_id = int(instrument_log_entry_form.log_entry_id.data)
                    log_entry = get_instrument_log_entry(log_entry_id)
                except ValueError:
                    return flask.abort(400)
                except InstrumentLogEntryDoesNotExistError:
                    return flask.abort(400)
                if instrument_log_entry_form.content_is_markdown.data:
                    log_entry_html = markdown_to_safe_html(instrument_log_entry_form.content.data)
                    mark_referenced_markdown_images_as_permanent(log_entry_html)
                if instrument_log_entry_form.has_event_utc_datetime.data:
                    event_utc_datetime = instrument_log_entry_form.event_utc_datetime.data
                else:
                    event_utc_datetime = None
                update_instrument_log_entry(
                    log_entry_id=log_entry_id,
                    content=instrument_log_entry_form.content.data,
                    category_ids=[
                        int(category_id) for category_id in instrument_log_entry_form.categories.data
                    ],
                    content_is_markdown=instrument_log_entry_form.content_is_markdown.data,
                    event_utc_datetime=event_utc_datetime
                )
                for file_storage in instrument_log_entry_form.files.data:
                    if file_storage.filename:
                        file_name = file_storage.filename
                        content = file_storage.stream.read()
                        create_instrument_log_file_attachment(log_entry.id, file_name, content)
                previously_attached_object_ids = {
                    object_attachment.object_id: object_attachment
                    for object_attachment in log_entry.object_attachments
                }
                for object_id in instrument_log_entry_form.objects.data:
                    if int(object_id) in previously_attached_object_ids:
                        object_attachment = previously_attached_object_ids[int(object_id)]
                        if object_attachment.is_hidden:
                            hide_instrument_log_object_attachment(object_attachment.id, set_hidden=False)
                        continue
                    try:
                        create_instrument_log_object_attachment(log_entry.id, int(object_id))
                    except ObjectDoesNotExistError:
                        continue
                if instrument_log_entry_form.existing_files.data:
                    for file_attachment_id in instrument_log_entry_form.existing_files.data:
                        file_attachment_id = int(file_attachment_id)
                        try:
                            file_attachment = get_instrument_log_file_attachment(file_attachment_id)
                            if file_attachment.log_entry_id == log_entry_id:
                                hide_instrument_log_file_attachment(file_attachment_id)
                        except InstrumentLogFileAttachmentDoesNotExistError:
                            continue
                if instrument_log_entry_form.existing_objects.data:
                    for object_attachment_id in instrument_log_entry_form.existing_objects.data:
                        object_attachment_id = int(object_attachment_id)
                        try:
                            object_attachment = get_instrument_log_object_attachment(object_attachment_id)
                            if object_attachment.log_entry_id == log_entry_id:
                                hide_instrument_log_object_attachment(object_attachment_id)
                        except InstrumentLogObjectAttachmentDoesNotExistError:
                            continue
                flask.flash(_('You have successfully updated the instrument log entry.'), 'success')
                return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
            else:
                return flask.abort(400)
        elif is_instrument_responsible_user or instrument.users_can_create_log_entries:
            log_entry_content = instrument_log_entry_form.content.data
            log_entry_has_files = any(file_storage.filename for file_storage in instrument_log_entry_form.files.data)
            object_ids = []
            for object_id in instrument_log_entry_form.objects.data:
                try:
                    object_ids.append(get_object(int(object_id)).id)
                except ObjectDoesNotExistError:
                    continue
                except ValueError:
                    continue
            log_entry_has_objects = bool(object_ids)
            if log_entry_content or log_entry_has_files or log_entry_has_objects:
                if instrument_log_entry_form.content_is_markdown.data:
                    log_entry_html = markdown_to_safe_html(instrument_log_entry_form.content.data)
                    mark_referenced_markdown_images_as_permanent(log_entry_html)
                if instrument_log_entry_form.has_event_utc_datetime.data:
                    event_utc_datetime = instrument_log_entry_form.event_utc_datetime.data
                else:
                    event_utc_datetime = None
                log_entry = create_instrument_log_entry(
                    instrument_id=instrument_id,
                    user_id=flask_login.current_user.id,
                    content=instrument_log_entry_form.content.data,
                    category_ids=[
                        int(category_id) for category_id in instrument_log_entry_form.categories.data
                    ],
                    content_is_markdown=instrument_log_entry_form.content_is_markdown.data,
                    event_utc_datetime=event_utc_datetime
                )
                for file_storage in instrument_log_entry_form.files.data:
                    if file_storage.filename:
                        file_name = file_storage.filename
                        content = file_storage.stream.read()
                        create_instrument_log_file_attachment(log_entry.id, file_name, content)
                for object_id in instrument_log_entry_form.objects.data:
                    try:
                        create_instrument_log_object_attachment(log_entry.id, int(object_id))
                    except ObjectDoesNotExistError:
                        continue
                flask.flash(_('You have created a new instrument log entry.'), 'success')
                return flask.redirect(flask.url_for(
                    '.instrument',
                    instrument_id=instrument_id,
                    _anchor=f'log_entry-{log_entry.id}'
                ))
            else:
                flask.flash(_('Please enter a log entry text, upload a file or select an object to create a log entry.'), 'error')
                return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
        else:
            flask.flash(_('You cannot create a log entry for this instrument.'), 'error')
            return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    elif instrument_log_entry_form.event_utc_datetime.errors:
        flask.flash(_("Please enter a valid event datetime."), 'error')
    user_settings = get_user_settings(flask_login.current_user.id)
    instrument_log_order_ascending = user_settings['INSTRUMENT_LOG_ORDER_ASCENDING']
    instrument_log_order_attribute = user_settings['INSTRUMENT_LOG_ORDER_ATTRIBUTE']
    if instrument_log_entries is not None:
        if instrument_log_order_attribute == 'datetime':
            instrument_log_entries.sort(
                key=lambda entry: entry.versions[-1].event_utc_datetime or entry.versions[0].utc_datetime,
                reverse=not instrument_log_order_ascending
            )
        elif instrument_log_order_attribute == 'username':
            instrument_log_entries.sort(
                key=lambda entry: (entry.author.name, entry.author.id),
                reverse=not instrument_log_order_ascending
            )

    instrument_actions = [
        action
        for action in get_actions(instrument_id=instrument.id)
        if Permissions.READ in get_user_action_permissions(action.id, flask_login.current_user.id) and (not action.is_hidden or flask_login.current_user.is_admin)
    ]
    instrument_actions.sort(key=lambda action: (action.id not in user_favorite_action_ids, get_translated_text(action.name), action.id))
    instrument_is_favorite = instrument.id in get_user_favorite_instrument_ids(flask_login.current_user.id)
    return flask.render_template(
        'instruments/instrument.html',
        instrument=instrument,
        instrument_is_favorite=instrument_is_favorite,
        instrument_actions=instrument_actions,
        instrument_log_entries=instrument_log_entries,
        instrument_log_users=instrument_log_users,
        instrument_log_categories=instrument_log_categories,
        attached_object_names=attached_object_names,
        is_instrument_responsible_user=is_instrument_responsible_user,
        instrument_log_entry_form=instrument_log_entry_form,
        instrument_log_order_form=InstrumentLogOrderForm(),
        mobile_upload_url=mobile_upload_url,
        mobile_upload_qrcode=mobile_upload_qrcode,
        instrument_log_order_ascending=instrument_log_order_ascending,
        instrument_log_order_attribute=instrument_log_order_attribute,
        linked_object=linked_object,
        object_link_form=object_link_form,
        linkable_action_ids=linkable_action_ids,
        already_linked_object_ids=already_linked_object_ids,
        show_object_title=show_object_title,
        languages=get_languages(),
        ActionType=ActionType,
        **template_kwargs
    )


class InstrumentForm(FlaskForm):
    instrument_responsible_users = SelectMultipleField()
    is_markdown = BooleanField(default=False)
    short_description_is_markdown = BooleanField(default=False)
    notes_is_markdown = BooleanField(default=False)
    translations = StringField(validators=[DataRequired()])
    categories = StringField(validators=[DataRequired()])
    users_can_create_log_entries = BooleanField(default=False)
    users_can_view_log_entries = BooleanField(default=False)
    create_log_entry_default = BooleanField(default=False)
    is_hidden = BooleanField(default=False)
    location = SelectField()
    show_linked_object_data = BooleanField(default=True)
    topics = SelectMultipleField()


@frontend.route('/instruments/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_instrument() -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    check_current_user_is_not_readonly()

    allowed_language_ids = [
        language.id
        for language in get_languages(only_enabled_for_input=True)
    ]

    instrument_form = InstrumentForm()
    instrument_form.instrument_responsible_users.choices = [
        (str(user.id), user)
        for user in get_users()
        if user.fed_id is None and (not user.is_hidden or (flask_login.current_user.is_admin and flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN']))
    ]
    instrument_form.topics.choices = [
        (str(topic.id), topic)
        for topic in get_topics()
    ]
    topic_ids_str = flask.request.args.get('topic_ids')
    if topic_ids_str is not None:
        valid_topic_id_strs = [
            topic_id_str
            for topic_id_str, topic in instrument_form.topics.choices
        ]
        instrument_form.topics.default = [
            topic_id_str
            for topic_id_str in topic_ids_str.split(',')
            if topic_id_str in valid_topic_id_strs
        ]
    all_choices, choices = get_locations_form_data(filter=lambda location: location.type is not None and location.type.enable_instruments)
    instrument_form.location.choices = choices
    instrument_form.location.all_choices = all_choices
    instrument_form.location.default = '-1'
    if not instrument_form.is_submitted():
        instrument_form.location.data = instrument_form.location.default

    if instrument_form.validate_on_submit():

        try:
            translation_data = json.loads(instrument_form.translations.data)
        except Exception:
            pass
        else:
            translation_keys = {'id', 'language_id', 'name', 'description',
                                'short_description', 'notes'}
            if not isinstance(translation_data, list):
                translation_data = ()
            for translation in translation_data:
                if not isinstance(translation, dict):
                    continue
                if set(translation.keys()) != translation_keys:
                    continue

                language_id = int(translation['language_id'])
                name = translation['name'].strip()

                if language_id == Language.ENGLISH:
                    try:
                        if len(name) < 1 or len(name) > 100:
                            flask.flash(_('Please enter an instrument name.'), 'error')
                            return flask.render_template('instruments/instrument_form.html',
                                                         submit_text='Create Instrument',
                                                         cancel_url=flask.url_for('.instruments'),
                                                         instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: typing.cast(int, t.value)),
                                                         ENGLISH=get_language(Language.ENGLISH),
                                                         languages=get_languages(only_enabled_for_input=True),
                                                         instrument_form=instrument_form
                                                         )
                    except Exception:
                        flask.flash(_('Please enter an instrument name.'), 'error')
                        return flask.render_template('instruments/instrument_form.html',
                                                     submit_text='Create Instrument',
                                                     cancel_url=flask.url_for('.instruments'),
                                                     instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: typing.cast(int, t.value)),
                                                     ENGLISH=get_language(Language.ENGLISH),
                                                     languages=get_languages(only_enabled_for_input=True),
                                                     instrument_form=instrument_form
                                                     )
            instrument = create_instrument(
                description_is_markdown=instrument_form.is_markdown.data,
                notes_is_markdown=instrument_form.notes_is_markdown.data,
                short_description_is_markdown=instrument_form.short_description_is_markdown.data,
                users_can_create_log_entries=instrument_form.users_can_create_log_entries.data,
                users_can_view_log_entries=instrument_form.users_can_view_log_entries.data,
                create_log_entry_default=instrument_form.create_log_entry_default.data,
                is_hidden=instrument_form.is_hidden.data,
                show_linked_object_data=instrument_form.show_linked_object_data.data,
            )

            for translation in translation_data:
                language_id = int(translation['language_id'])
                name = translation['name'].strip()
                description = translation['description'].strip()
                short_description = translation['short_description'].strip()
                notes = translation['notes'].strip()

                if language_id not in allowed_language_ids:
                    continue

                if instrument_form.is_markdown.data:
                    description_as_html = markdown_to_safe_html(description, anchor_prefix="instrument-description")
                    mark_referenced_markdown_images_as_permanent(description_as_html)
                if instrument_form.notes_is_markdown.data:
                    notes_as_html = markdown_to_safe_html(notes, anchor_prefix="instrument-notes")
                    mark_referenced_markdown_images_as_permanent(notes_as_html)
                if instrument_form.short_description_is_markdown.data:
                    short_description_as_html = markdown_to_safe_html(short_description, anchor_prefix="instrument-short-description")
                    mark_referenced_markdown_images_as_permanent(short_description_as_html)

                set_instrument_translation(
                    instrument_id=instrument.id,
                    language_id=language_id,
                    name=name,
                    description=description,
                    short_description=short_description,
                    notes=notes
                )

        flask.flash(_('The instrument was created successfully.'), 'success')
        instrument_responsible_user_ids = [
            int(user_id)
            for user_id in instrument_form.instrument_responsible_users.data
        ]
        set_instrument_responsible_users(instrument.id, instrument_responsible_user_ids)
        if not flask.current_app.config['DISABLE_TOPICS']:
            topic_ids = [
                int(topic_id)
                for topic_id in instrument_form.topics.data
            ]
            set_instrument_topics(instrument.id, topic_ids)

        if instrument_form.location.data is not None and instrument_form.location.data != '-1':
            set_instrument_location(instrument.id, int(instrument_form.location.data))

        try:
            category_data = json.loads(instrument_form.categories.data)
        except Exception:
            pass
        else:
            themes_by_name = {
                theme.name.lower(): theme
                for theme in list(InstrumentLogCategoryTheme)
            }
            category_keys = {'id', 'title', 'theme'}
            if not isinstance(category_data, list):
                category_data = ()
            for category in category_data:
                # skip any invalid entries
                if not isinstance(category, dict):
                    continue
                if set(category.keys()) != category_keys:
                    continue
                if not all(isinstance(category[key], str) for key in category_keys):
                    continue
                category_title = category['title'].strip()
                if not category_title:
                    continue
                if len(category_title) > 100:
                    continue
                try:
                    category_id = int(category['id'])
                except ValueError:
                    continue
                category_theme = themes_by_name.get(category['theme'], InstrumentLogCategoryTheme.GRAY)
                if category_id < 0:
                    create_instrument_log_category(
                        instrument.id,
                        category['title'],
                        category_theme
                    )
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument.id))
    return flask.render_template(
        'instruments/instrument_form.html',
        submit_text='Create Instrument',
        cancel_url=flask.url_for('.instruments'),
        instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: typing.cast(int, t.value)),
        ENGLISH=get_language(Language.ENGLISH),
        languages=get_languages(only_enabled_for_input=True),
        instrument_form=instrument_form,
        instrument_translations={}
    )


@frontend.route('/instruments/<int:instrument_id>/edit', methods=['GET', 'POST'])
@flask_login.login_required
def edit_instrument(instrument_id: int) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    if instrument.component_id is not None:
        flask.flash(_('Editing imported instruments is not yet supported.'), 'error')
        return flask.abort(403)
    check_current_user_is_not_readonly()
    if not flask_login.current_user.is_admin and flask_login.current_user not in instrument.responsible_users:
        return flask.abort(403)

    allowed_language_ids = [
        language.id
        for language in get_languages(only_enabled_for_input=True)
    ]

    instrument_form = InstrumentForm()

    instrument_form.instrument_responsible_users.choices = [
        (str(user.id), user)
        for user in get_users()
        if user in instrument.responsible_users or (user.fed_id is None and (not user.is_hidden or (flask_login.current_user.is_admin and flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'])))
    ]
    instrument_form.instrument_responsible_users.default = [
        str(user.id)
        for user in instrument.responsible_users
    ]
    instrument_form.topics.choices = [
        (str(topic.id), topic)
        for topic in get_topics()
    ]
    instrument_form.topics.default = [
        str(topic.id)
        for topic in instrument.topics
    ]
    all_choices, choices = get_locations_form_data(filter=lambda location: location.type is not None and location.type.enable_instruments)
    instrument_form.location.choices = choices
    instrument_form.location.all_choices = all_choices

    if instrument.location_id is not None:
        instrument_form.location.default = str(instrument.location_id)
    else:
        instrument_form.location.default = '-1'

    if not instrument_form.is_submitted():
        instrument_form.is_markdown.data = instrument.description_is_markdown
        instrument_form.short_description_is_markdown.data = instrument.short_description_is_markdown
        instrument_form.notes_is_markdown.data = instrument.notes_is_markdown
        instrument_form.users_can_create_log_entries.data = instrument.users_can_create_log_entries
        instrument_form.users_can_view_log_entries.data = instrument.users_can_view_log_entries
        instrument_form.create_log_entry_default.data = instrument.create_log_entry_default
        instrument_form.is_hidden.data = instrument.is_hidden
        instrument_form.location.data = instrument_form.location.default
        instrument_form.show_linked_object_data.data = instrument.show_linked_object_data
    location_is_invalid = instrument_form.location.data not in {
        location_id_str
        for location_id_str, location_name in choices
    }
    if instrument_form.validate_on_submit():
        update_instrument(
            instrument_id=instrument.id,
            description_is_markdown=instrument_form.is_markdown.data,
            short_description_is_markdown=instrument_form.short_description_is_markdown.data,
            notes_is_markdown=instrument_form.notes_is_markdown.data,
            users_can_create_log_entries=instrument_form.users_can_create_log_entries.data,
            users_can_view_log_entries=instrument_form.users_can_view_log_entries.data,
            create_log_entry_default=instrument_form.create_log_entry_default.data,
            is_hidden=instrument_form.is_hidden.data,
            show_linked_object_data=instrument_form.show_linked_object_data.data
        )
        instrument_responsible_user_ids = [
            int(user_id)
            for user_id in instrument_form.instrument_responsible_users.data
        ]
        set_instrument_responsible_users(instrument.id, instrument_responsible_user_ids)
        if not flask.current_app.config['DISABLE_TOPICS']:
            topic_ids = [
                int(topic_id)
                for topic_id in instrument_form.topics.data
            ]
            set_instrument_topics(instrument.id, topic_ids)

        if instrument_form.location.data is not None and instrument_form.location.data != '-1':
            set_instrument_location(instrument.id, int(instrument_form.location.data))
        else:
            set_instrument_location(instrument.id, None)

        # translations
        try:
            translation_data = json.loads(instrument_form.translations.data)
        except Exception:
            pass
        else:
            translation_keys = {'language_id', 'name', 'description',
                                'short_description', 'notes'}

            if not isinstance(translation_data, list):
                translation_data = ()

            valid_translations = set()
            for translation in translation_data:
                if not isinstance(translation, dict):
                    continue
                if set(translation.keys()) != translation_keys:
                    continue
                language_id = int(translation['language_id'])
                name = translation['name'].strip()
                description = translation['description'].strip()
                short_description = translation['short_description']
                notes = translation['notes'].strip()

                if language_id not in allowed_language_ids:
                    continue

                if language_id == Language.ENGLISH:
                    error = False
                    if not isinstance(name, str) or not isinstance(description, str)\
                            or not isinstance(short_description, str) or not isinstance(notes, str):
                        error = True
                    if len(name) < 1 or len(name) > 100:
                        error = True
                    if error:
                        flask.flash(_('Please enter an english instrument name.'), 'error')
                        instrument_translations_list = get_instrument_translations_for_instrument(instrument.id)
                        instrument_language_ids = [
                            translation.language_id
                            for translation in instrument_translations_list

                        ]
                        instrument_translations = {
                            instrument_translation.language_id: instrument_translation
                            for instrument_translation in instrument_translations_list
                        }
                        ENGLISH = get_language(Language.ENGLISH)
                        return flask.render_template(
                            'instruments/instrument_form.html',
                            submit_text=_('Save'),
                            cancel_url=flask.url_for('.instrument', instrument_id=instrument_id),
                            instrument_log_category_themes=sorted(
                                InstrumentLogCategoryTheme,
                                key=lambda t: int(t.value)
                            ),
                            instrument_translations=instrument_translations,
                            instrument_language_ids=instrument_language_ids,
                            ENGLISH=ENGLISH,
                            instrument_log_categories=get_instrument_log_categories(instrument.id),
                            languages=get_languages(only_enabled_for_input=True),
                            instrument_form=instrument_form
                        )

                if len(name + description + short_description + notes) == 0:
                    continue

                if instrument_form.is_markdown.data:
                    description_as_html = markdown_to_safe_html(description, anchor_prefix='instrument-description')
                    mark_referenced_markdown_images_as_permanent(description_as_html)

                if instrument_form.short_description_is_markdown.data:
                    short_description_as_html = markdown_to_safe_html(short_description, anchor_prefix='instrument-short-description')
                    mark_referenced_markdown_images_as_permanent(short_description_as_html)
                if instrument_form.notes_is_markdown.data:
                    notes_as_html = markdown_to_safe_html(notes, anchor_prefix='instrument-notes')
                    mark_referenced_markdown_images_as_permanent(notes_as_html)

                new_translation = set_instrument_translation(
                    instrument_id=instrument.id,
                    language_id=language_id,
                    name=name,
                    description=description,
                    notes=notes,
                    short_description=short_description,
                )
                valid_translations.add((new_translation.instrument_id, new_translation.language_id))

            for translation in get_instrument_translations_for_instrument(instrument_id):
                if (translation.instrument_id, translation.language_id) not in valid_translations:
                    delete_instrument_translation(
                        instrument_id=translation.instrument_id,
                        language_id=translation.language_id
                    )

        try:
            category_data = json.loads(instrument_form.categories.data)
        except Exception:
            pass
        else:
            themes_by_name = {
                theme.name.lower(): theme
                for theme in list(InstrumentLogCategoryTheme)
            }
            category_keys = {'id', 'title', 'theme'}
            if not isinstance(category_data, list):
                category_data = ()
            existing_category_ids = {
                category.id
                for category in get_instrument_log_categories(instrument_id)
            }
            valid_category_ids = set()
            for category in category_data:
                # skip any invalid entries
                if not isinstance(category, dict):
                    continue
                if set(category.keys()) != category_keys:
                    continue
                if not all(isinstance(category[key], str) for key in category_keys):
                    continue
                category_title = category['title'].strip()
                if not category_title:
                    continue
                if len(category_title) > 100:
                    continue
                try:
                    category_id = int(category['id'])
                except ValueError:
                    continue
                category_theme = themes_by_name.get(category['theme'], InstrumentLogCategoryTheme.GRAY)
                if category_id < 0:
                    new_category = create_instrument_log_category(
                        instrument.id,
                        category['title'],
                        category_theme
                    )
                    valid_category_ids.add(new_category.id)
                elif category_id in existing_category_ids:
                    update_instrument_log_category(
                        category_id,
                        category['title'],
                        category_theme
                    )
                    valid_category_ids.add(category_id)
            for category in get_instrument_log_categories(instrument_id):
                if category.id not in valid_category_ids:
                    delete_instrument_log_category(category.id)
        flask.flash(_('The instrument was updated successfully.'), 'success')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument.id))

    instrument_translations_list = get_instrument_translations_for_instrument(instrument.id)
    instrument_language_ids = [translation.language_id for translation in instrument_translations_list]
    instrument_translations = {
        instrument_translation.language_id: instrument_translation
        for instrument_translation in instrument_translations_list
    }
    english = get_language(Language.ENGLISH)
    return flask.render_template(
        'instruments/instrument_form.html',
        submit_text=_('Save'),
        cancel_url=flask.url_for('.instrument', instrument_id=instrument_id),
        instrument_log_category_themes=sorted(InstrumentLogCategoryTheme, key=lambda t: int(t.value)),
        instrument_translations=instrument_translations,
        instrument_language_ids=instrument_language_ids,
        ENGLISH=english,
        instrument_log_categories=get_instrument_log_categories(instrument.id),
        languages=get_languages(only_enabled_for_input=True),
        instrument_form=instrument_form,
        location_is_invalid=location_is_invalid
    )


@frontend.route('/instruments/<int:instrument_id>/log/<int:log_entry_id>/file_attachments/<int:file_attachment_id>')
@flask_login.login_required
def instrument_log_file_attachment(instrument_id: int, log_entry_id: int, file_attachment_id: int) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    if not instrument.users_can_view_log_entries and flask_login.current_user not in instrument.responsible_users:
        return flask.abort(403)
    try:
        file_attachment = get_instrument_log_file_attachment(file_attachment_id)
    except InstrumentLogFileAttachmentDoesNotExistError:
        return flask.abort(404)
    return_preview = 'preview' in flask.request.args
    return_thumbnail = 'thumbnail' in flask.request.args

    if return_thumbnail:
        if file_attachment.image_info:
            return flask.Response(
                file_attachment.image_info.thumbnail_content,
                200,
                headers={
                    'Content-Type': file_attachment.image_info.thumbnail_mime_type
                }
            )
        else:
            return_preview = True
    if return_preview:
        file_extension = os.path.splitext(file_attachment.file_name)[1]
        mime_type = flask.current_app.config.get('MIME_TYPES', {}).get(file_extension, '')
        if mime_type:
            return flask.Response(
                file_attachment.content,
                200,
                headers={
                    'Content-Type': mime_type
                }
            )
    return flask.Response(
        file_attachment.content,
        200,
        headers={
            'Content-Disposition': f'attachment; filename="{file_attachment.file_name}"'
        }
    )


@frontend.route('/instruments/<int:instrument_id>/log/mobile_upload/<token>', methods=['GET'])
def instrument_log_mobile_file_upload(instrument_id: int, token: str) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
    try:
        user_id, instrument_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if not instrument.users_can_create_log_entries and user not in instrument.responsible_users:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    return flask.render_template('mobile_upload.html', user_id=get_user(user_id), instrument=instrument)


@frontend.route('/instruments/<int:instrument_id>/log/mobile_upload/<token>', methods=['POST'])
def post_instrument_log_mobile_file_upload(instrument_id: int, token: str) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)
    serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='instrument-log-mobile-upload')
    try:
        user_id, instrument_id = serializer.loads(token, max_age=15 * 60)
    except itsdangerous.BadSignature:
        return flask.abort(400)
    try:
        user = get_user(user_id)
    except UserDoesNotExistError:
        return flask.abort(403)
    if not instrument.users_can_create_log_entries and user not in instrument.responsible_users:
        return flask.abort(403)
    if user.is_readonly:
        return flask.abort(403)
    files = flask.request.files.getlist('file_input')
    if not files:
        return flask.redirect(
            flask.url_for(
                '.instrument_log_mobile_file_upload',
                instrument_id=instrument_id,
                token=token
            )
        )
    if files:
        log_entry = create_instrument_log_entry(
            instrument_id=instrument_id,
            user_id=user_id,
            content='',
            category_ids=[]
        )
        for file_storage in files:
            create_instrument_log_file_attachment(
                instrument_log_entry_id=log_entry.id,
                file_name=file_storage.filename or '',
                content=file_storage.stream.read()
            )
    return flask.render_template('mobile_upload_success.html')


@frontend.route('/users/me/settings/instrument_log_order', methods=['POST'])
@flask_login.login_required
def set_instrument_log_order() -> FlaskResponseT:
    form = InstrumentLogOrderForm()
    if not form.validate_on_submit():
        return flask.abort(400)
    ascending = form.ascending.data
    attribute = form.attribute.data
    set_user_settings(flask_login.current_user.id, {
        'INSTRUMENT_LOG_ORDER_ASCENDING': ascending,
        'INSTRUMENT_LOG_ORDER_ATTRIBUTE': attribute
    })
    return "", 200


@frontend.route('/instruments/<int:instrument_id>/link_object', methods=['GET', 'POST'])
@flask_login.login_required
def instrument_link_object(instrument_id: int) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)

    is_instrument_responsible_user = any(
        responsible_user.id == flask_login.current_user.id
        for responsible_user in instrument.responsible_users
    )
    if not is_instrument_responsible_user:
        return flask.abort(403)
    check_current_user_is_not_readonly()
    object_link_form = ObjectLinkForm()
    if not object_link_form.validate_on_submit():
        flask.flash(_("Missing or invalid object ID."), 'error')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    object_id = object_link_form.object_id.data
    try:
        if Permissions.GRANT not in get_user_object_permissions(object_id, flask_login.current_user.id) and not flask_login.current_user.has_admin_permissions:
            flask.flash(_("You do not have GRANT permissions for this object."), 'error')
            return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
        object = get_object(object_id)
        object_is_valid = False
        if object.action_id is not None:
            action = get_action(object.action_id)
            if action.type_id is not None:
                action_type = get_action_type(action.type_id)
                object_is_valid = action_type.enable_instrument_link
        if not object_is_valid:
            flask.flash(_("This object cannot be linked to an instrument."), 'error')
            return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
        set_instrument_object(instrument_id, object_id)
    except ObjectDoesNotExistError:
        flask.flash(_("Object does not exist."), 'error')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    except InstrumentObjectLinkAlreadyExistsError:
        flask.flash(_("The instrument is already linked to an object or the object is already linked to an instrument."), 'error')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    flask.flash(_("Successfully linked the object to the instrument."), 'success')
    return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))


@frontend.route('/instruments/<int:instrument_id>/unlink_object', methods=['GET', 'POST'])
@flask_login.login_required
def instrument_unlink_object(instrument_id: int) -> FlaskResponseT:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        return flask.abort(404)
    try:
        instrument = get_instrument(instrument_id)
    except InstrumentDoesNotExistError:
        return flask.abort(404)

    is_instrument_responsible_user = any(
        responsible_user.id == flask_login.current_user.id
        for responsible_user in instrument.responsible_users
    )
    if not is_instrument_responsible_user:
        return flask.abort(403)
    check_current_user_is_not_readonly()
    object_link_form = ObjectLinkForm()
    if not object_link_form.validate_on_submit():
        flask.flash(_("Missing or invalid object ID."), 'error')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    object_id = object_link_form.object_id.data
    if instrument.object_id != object_id:
        flask.flash(_('This instrument is not linked to this object.'), 'error')
        return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))
    set_instrument_object(instrument_id=instrument_id, object_id=None)
    flask.flash(_("Successfully unlinked the object from the instrument."), 'success')
    return flask.redirect(flask.url_for('.instrument', instrument_id=instrument_id))



# coding: utf-8
"""
Add missing NOT NULL constraints to columns in languages table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('languages', 'datetime_format_datetime') and not column_is_nullable('languages', 'datetime_format_moment'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE languages
            ALTER COLUMN datetime_format_datetime SET NOT NULL,
            ALTER COLUMN datetime_format_moment SET NOT NULL
    """))
    return True



# coding: utf-8
"""
Add location_id column to instruments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('instruments', 'location_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE instruments
            ADD location_id INTEGER,
            ADD FOREIGN KEY(location_id) REFERENCES locations(id)
    """))
    return True



# coding: utf-8
"""
Add description_is_markdown and short_description_is_markdown columns to topics table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    has_description_is_markdown = table_has_column('topics', 'description_is_markdown')
    has_short_description_is_markdown = table_has_column('topics', 'short_description_is_markdown')
    # Skip migration by condition
    if has_description_is_markdown and has_short_description_is_markdown:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE topics
        ADD description_is_markdown BOOLEAN DEFAULT FALSE NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE topics
        ALTER COLUMN description_is_markdown DROP DEFAULT
    """))
    db.session.execute(db.text("""
        ALTER TABLE topics
        ADD short_description_is_markdown BOOLEAN DEFAULT FALSE NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE topics
        ALTER COLUMN short_description_is_markdown DROP DEFAULT
    """))
    return True



# coding: utf-8
"""

"""
import datetime
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .permissions import Permissions
from .groups import Group
from .users import User
from .utils import Model

if typing.TYPE_CHECKING:
    from .group_categories import GroupCategory


class Project(Model):
    __tablename__ = 'projects'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    name: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    description: Mapped[typing.Dict[str, str]] = db.Column(postgresql.JSON, nullable=False)
    categories: Mapped[typing.List['GroupCategory']] = relationship('GroupCategory', secondary='project_group_category_association_table', back_populates='project_groups')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Project"]]


class UserProjectPermissions(Model):
    __tablename__ = 'user_project_permissions'

    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserProjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(project_id, user_id),
    )


class GroupProjectPermissions(Model):
    __tablename__ = 'group_project_permissions'

    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupProjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(project_id, group_id),
    )


class SubprojectRelationship(Model):
    __tablename__ = 'subproject_relationship'

    parent_project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    child_project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    child_can_add_users_to_parent: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["SubprojectRelationship"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(parent_project_id, child_project_id),
    )


class ProjectInvitation(Model):
    __tablename__ = 'project_invitations'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    inviter_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    accepted: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    revoked: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())  # TODO: migration

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ProjectInvitation"]]


class ProjectObjectAssociation(Model):
    __tablename__ = 'project_object_association'

    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Project.id, ondelete="CASCADE"), primary_key=True)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('objects_current.object_id', ondelete="CASCADE"), nullable=False, unique=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ProjectObjectAssociation"]]



import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_str, _get_dict
from ..users import get_mutable_user, create_user, set_user_hidden, get_user, get_user_alias, User
from ..components import Component
from .. import errors, fed_logs
from ...models import UserType
from ... import db


class UserData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    name: typing.Optional[str]
    email: typing.Optional[str]
    orcid: typing.Optional[str]
    affiliation: typing.Optional[str]
    role: typing.Optional[str]
    extra_fields: typing.Dict[str, str]


class UserRef(typing.TypedDict):
    user_id: int
    component_uuid: str


class SharedUserData(typing.TypedDict):
    user_id: int
    component_uuid: str
    name: typing.Optional[str]
    email: typing.Optional[str]
    orcid: typing.Optional[str]
    affiliation: typing.Optional[str]
    role: typing.Optional[str]
    extra_fields: typing.Dict[str, str]


def import_user(
        user_data: UserData,
        component: Component
) -> User:
    component_id = _get_or_create_component_id(user_data['component_uuid'])
    # component_id will only be None if this would import a local user
    assert component_id is not None
    try:
        mutable_user = get_mutable_user(user_data['fed_id'], component_id)
        ignored_keys = {
            'fed_id',
            'component_uuid'
        }
        if any(
                value != getattr(mutable_user, key)
                for key, value in user_data.items()
                if key not in ignored_keys
        ):
            mutable_user.name = user_data['name']
            mutable_user.email = user_data['email']
            mutable_user.orcid = user_data['orcid']
            mutable_user.affiliation = user_data['affiliation']
            mutable_user.role = user_data['role']
            mutable_user.extra_fields = user_data['extra_fields']
            db.session.add(mutable_user)
            db.session.commit()
            fed_logs.update_user(mutable_user.id, component.id)
        user = User.from_database(mutable_user)
    except errors.UserDoesNotExistError:
        user = create_user(
            fed_id=user_data['fed_id'],
            component_id=component_id,
            name=user_data['name'],
            email=user_data['email'],
            orcid=user_data['orcid'],
            affiliation=user_data['affiliation'],
            role=user_data['role'],
            extra_fields=user_data['extra_fields'],
            type=UserType.FEDERATION_USER
        )
        set_user_hidden(user.id, True)
        fed_logs.import_user(user.id, component.id)
    return user


def parse_import_user(
        user_data: typing.Dict[str, typing.Any],
        component: Component
) -> User:
    return import_user(parse_user(user_data, component), component)


def parse_user(
        user_data: typing.Dict[str, typing.Any],
        component: Component
) -> UserData:
    uuid = _get_uuid(user_data.get('component_uuid'))
    fed_id = _get_id(user_data.get('user_id'))
    if uuid != component.uuid:
        # only accept user data from original source
        raise errors.InvalidDataExportError(f'User data update for user #{fed_id} @ {uuid}')
    return UserData(
        fed_id=fed_id,
        component_uuid=uuid,
        name=_get_str(user_data.get('name')),
        email=_get_str(user_data.get('email')),
        orcid=_get_str(user_data.get('orcid')),
        affiliation=_get_str(user_data.get('affiliation')),
        role=_get_str(user_data.get('role')),
        extra_fields=_get_dict(user_data.get('extra_fields'), default={})
    )


@typing.overload
def _parse_user_ref(
        user_data: typing.Union[UserRef, typing.Dict[str, typing.Any]]
) -> UserRef:
    ...


@typing.overload
def _parse_user_ref(
        user_data: None
) -> None:
    ...


def _parse_user_ref(
        user_data: typing.Optional[typing.Union[UserRef, typing.Dict[str, typing.Any]]]
) -> typing.Optional[UserRef]:
    if user_data is None:
        return None
    user_id = _get_id(user_data.get('user_id'))
    component_uuid = _get_uuid(user_data.get('component_uuid'))
    return UserRef(
        user_id=user_id,
        component_uuid=component_uuid
    )


@typing.overload
def _get_or_create_user_id(
        user_data: UserRef
) -> int:
    ...


@typing.overload
def _get_or_create_user_id(
        user_data: None
) -> None:
    ...


def _get_or_create_user_id(
        user_data: typing.Optional[UserRef]
) -> typing.Optional[int]:
    if user_data is None:
        return None
    component_id = _get_or_create_component_id(user_data['component_uuid'])
    try:
        user = get_user(user_data['user_id'], component_id)
    except errors.UserDoesNotExistError:
        assert component_id is not None
        user = create_user(
            name=None,
            email=None,
            fed_id=user_data['user_id'],
            component_id=component_id,
            type=UserType.FEDERATION_USER
        )
        set_user_hidden(user.id, True)
        fed_logs.create_ref_user(user.id, component_id)
    return user.id


def shared_user_preprocessor(
        user_id: int,
        component: Component,
        _refs: typing.List[typing.Tuple[str, int]],
        _markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedUserData]:
    user = get_user(user_id)
    if user.component_id is not None:
        return None
    try:
        alias = get_user_alias(user_id, component.id)
        return SharedUserData(
            user_id=user.id,
            component_uuid=flask.current_app.config['FEDERATION_UUID'],
            name=alias.name,
            email=alias.email,
            orcid=alias.orcid,
            affiliation=alias.affiliation,
            role=alias.role,
            extra_fields=alias.extra_fields
        )
    except errors.UserAliasDoesNotExistError:
        return SharedUserData(
            user_id=user.id,
            component_uuid=flask.current_app.config['FEDERATION_UUID'],
            name=None,
            email=None,
            orcid=None,
            affiliation=None,
            role=None,
            extra_fields={}
        )



# coding: utf-8
"""
Script for creating an action in SampleDB.

Usage: sampledb create_action <instrument_id> <type: sample or measurement > <name> <description> <schema_file_name>
"""

import json
import sys
import typing

from .. import create_app
from ..logic.action_types import check_action_type_exists, get_action_type, get_action_types
from ..logic.actions import create_action
from ..logic.action_translations import set_action_translation
from ..logic.instruments import check_instrument_exists
from ..logic.languages import Language
from ..logic.schemas import validate_schema
from ..logic.errors import InstrumentDoesNotExistError, ValidationError, ActionTypeDoesNotExistError
from ..logic.utils import get_translated_text
from .. import models


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 5:
        print(__doc__)
        sys.exit(1)
    instrument_id_str, action_type_id_str, name, description, schema_file_name = arguments
    if instrument_id_str == 'None':
        instrument_id: typing.Optional[int] = None
    else:
        try:
            instrument_id = int(instrument_id_str)
        except ValueError:
            print("Error: instrument_id must be an integer or 'None'", file=sys.stderr)
            sys.exit(1)
    try:
        action_type_id: typing.Optional[int] = int(action_type_id_str)
    except ValueError:
        action_type_id = {
            'sample': models.ActionType.SAMPLE_CREATION,
            'measurement': models.ActionType.MEASUREMENT,
            'simulation': models.ActionType.SIMULATION
        }.get(action_type_id_str, None)
    if action_type_id is not None:
        try:
            check_action_type_exists(action_type_id)
        except ActionTypeDoesNotExistError:
            action_type_id = None
    if action_type_id is None:
        print("Error: action type must be one of the following:", file=sys.stderr)
        print(f'- "sample" (for {get_translated_text(get_action_type(models.ActionType.SAMPLE_CREATION).name, "en")})', file=sys.stderr)
        print(f'- "measurement" (for {get_translated_text(get_action_type(models.ActionType.MEASUREMENT).name, "en")})', file=sys.stderr)
        print(f'- "simulation" (for {get_translated_text(get_action_type(models.ActionType.SIMULATION).name, "en")})', file=sys.stderr)
        for action_type in get_action_types():
            print(f'- "{action_type.id}" (for {get_translated_text(action_type.name, "en")})', file=sys.stderr)
        sys.exit(1)
    app = create_app()
    with app.app_context():
        if instrument_id is not None:
            if app.config['DISABLE_INSTRUMENTS']:
                print('Error: instruments are disabled', file=sys.stderr)
                sys.exit(1)
            try:
                check_instrument_exists(instrument_id)
            except InstrumentDoesNotExistError:
                print('Error: no instrument with this id exists', file=sys.stderr)
                sys.exit(1)
        with open(schema_file_name, 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        try:
            validate_schema(schema)
        except ValidationError as e:
            print(f'Error: invalid schema: {str(e)}', file=sys.stderr)
            sys.exit(1)
        action = create_action(
            instrument_id=instrument_id,
            action_type_id=action_type_id,
            schema=schema
        )
        set_action_translation(
            language_id=Language.ENGLISH,
            action_id=action.id,
            name=name,
            description=description
        )
        print(f"Success: the action has been created in SampleDB (#{action.id})")



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .groups import Group
from .users import User
from .permissions import Permissions
from .utils import Model


class DefaultUserPermissions(Model):
    __tablename__ = 'default_user_permissions'

    creator_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["DefaultUserPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, user_id),
    )


class DefaultGroupPermissions(Model):
    __tablename__ = 'default_group_permissions'

    creator_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["DefaultGroupPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, group_id),
    )


class DefaultProjectPermissions(Model):
    __tablename__ = 'default_project_permissions'

    creator_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["DefaultProjectPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(creator_id, project_id),
    )


class AllUserDefaultPermissions(Model):
    __tablename__ = 'all_user_default_permissions'

    creator_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), primary_key=True)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["AllUserDefaultPermissions"]]



# coding: utf-8
"""

"""
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


class InfoPage(Model):
    __tablename__ = 'info_pages'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    title: Mapped[typing.Dict[str, str]] = db.Column(db.JSON, nullable=False)
    content: Mapped[typing.Dict[str, str]] = db.Column(db.JSON, nullable=False)
    endpoint: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    disabled: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)

    acknowledgements = db.relationship(
        'InfoPageAcknowledgement',
        back_populates='info_page',
        cascade='all, delete',
        passive_deletes=True
    )

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InfoPage"]]

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, title={self.title})>'


class InfoPageAcknowledgement(Model):
    __tablename__ = 'info_page_acknowledgements'

    info_page_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('info_pages.id', ondelete='CASCADE'), primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), primary_key=True)
    utc_datetime: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.DateTime, nullable=True)

    info_page = db.relationship(
        'InfoPage',
        back_populates='acknowledgements'
    )

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InfoPageAcknowledgement"]]

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(info_page_id={self.info_page_id}, user_id={self.user_id})>'



# coding: utf-8
"""
Replace NOT NULL constraints by not-empty-check in action_translations.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    constraints = db.session.execute(db.text("""
         SELECT conname
         FROM pg_catalog.pg_constraint
         WHERE conname = 'action_translations_not_empty_check'
    """)).fetchall()

    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
            ALTER TABLE action_translations
                ADD CONSTRAINT action_translations_not_empty_check
                    CHECK (NOT (name IS NULL AND description IS NULL AND short_description IS NULL)),
                ALTER COLUMN name DROP NOT NULL,
                ALTER COLUMN description DROP NOT NULL,
                ALTER COLUMN short_description DROP NOT NULL
        """))
    return True



"""
Background Tasks

Some tasks, such as sending mails or sending HTTP requests to other servers,
may need to be performed asynchronously to prevent the frontend from being
blocked for their duration. This module allows posting tasks to be handled
in the background, as well as querying tasks to check their status or results.

Currently, background tasks need to be enabled via the ENABLE_BACKGROUND_TASKS
configuration value. If they are not enabled, tasks will be performed
synchronously instead.
"""

import sys
import threading
import traceback
import typing

import flask

from ... import db
from ...utils import ansi_color
from ...models import BackgroundTask, BackgroundTaskStatus
from .. import errors
from .background_dataverse_export import handle_dataverse_export_task
from .automatic_schema_updates import handle_check_for_automatic_schema_updates_task, handle_perform_automatic_schema_updates_task
from .send_mail import handle_send_mail_task
from .poke_components import handle_poke_components_task
from .trigger_webhooks import handle_trigger_object_log_webhooks, handle_webhook_send

TASK_WAIT_TIMEOUT = 30
NUM_HANDLER_THREADS = 4

HANDLERS: typing.Dict[str, typing.Callable[[typing.Dict[str, typing.Any], typing.Optional[int]], typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]]] = {
    'send_mail': handle_send_mail_task,
    'dataverse_export': handle_dataverse_export_task,
    'poke_components': handle_poke_components_task,
    'trigger_object_log_webhooks': handle_trigger_object_log_webhooks,
    'webhook_send': handle_webhook_send,
    'check_for_automatic_schema_updates': handle_check_for_automatic_schema_updates_task,
    'perform_automatic_schema_updates': handle_perform_automatic_schema_updates_task,
}

should_stop = False
wake_event = threading.Event()

handler_threads: typing.List[threading.Thread] = []
cleanup_thread: typing.Optional[threading.Thread] = None


def get_background_tasks() -> typing.Sequence[BackgroundTask]:
    return typing.cast(typing.Sequence[BackgroundTask], BackgroundTask.query.order_by(BackgroundTask.id).all())


def get_background_task(task_id: int) -> BackgroundTask:
    task: typing.Optional[BackgroundTask] = BackgroundTask.query.filter_by(id=task_id).first()
    if task is None:
        raise errors.BackgroundTaskDoesNotExistError()
    return task


def post_background_task(
        type: str,
        data: typing.Dict[str, typing.Any],
        auto_delete: bool = True
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    """
    Create a background task and post it to be performed.

    If background tasks are disabled, the task will be performed synchronously
    instead and no BackgroundTask object will be returned. If a task does not
    have the auto_delete flag set, its ID should be stored so that the result
    can be queried later and so that it can be deleted.

    :param type: the type of the task
    :param data: data for the task
    :param auto_delete: whether the task should be deleted automatically, once
        it is done or has failed
    :return: the task status and the task object itself
    """
    if flask.current_app.config['ENABLE_BACKGROUND_TASKS']:
        if flask.current_app.config['TESTING']:
            auto_delete = False
        task = BackgroundTask(
            type=type,
            auto_delete=auto_delete,
            data=data,
            status=BackgroundTaskStatus.POSTED
        )
        db.session.add(task)
        db.session.commit()
        wake_event.set()
        start_handler_threads(flask.current_app)
        return BackgroundTaskStatus.POSTED, task
    else:
        return _handle_background_task(type, data, None)[0], None


def start_handler_threads(app: flask.Flask) -> None:
    """
    Start handler threads for background tasks.

    This function first cleans up all dead threads, then creates and starts
    handler threads until the desired number of handler threads has been
    reached.

    If background tasks are disabled, this function returns immediately.
    """
    if not app.config['ENABLE_BACKGROUND_TASKS']:
        return

    # remove handler threads that might have died
    for handler_thread in handler_threads.copy():
        if not handler_thread.is_alive():
            handler_threads.remove(handler_thread)

    # get actual app instead of thread local proxy from this thread to pass it to the new thread
    get_current_app = getattr(app, '_get_current_object', None)
    if get_current_app is not None:
        app = get_current_app()
    # use daemon threads during testing, as a failed test may circumvent the thread stop signal
    daemon = app.config.get('TESTING', False)
    while len(handler_threads) < NUM_HANDLER_THREADS:
        handler_thread = threading.Thread(target=_handle_background_tasks, args=[app, len(handler_threads) == 0], daemon=daemon)
        handler_thread.start()
        handler_threads.append(handler_thread)

    if not daemon:
        global cleanup_thread
        # create thread that takes care of stopping the background task threads if
        # the main thread exits without stopping them, e.g. for scripts
        # this is only necessary if the handler threads are not daemons
        if cleanup_thread is not None and not cleanup_thread.is_alive():
            cleanup_thread.join()
            cleanup_thread = None
        if cleanup_thread is None:
            def cleanup_thread_function() -> None:
                threading.main_thread().join()
                stop_handler_threads(app)

            cleanup_thread = threading.Thread(target=cleanup_thread_function)
            cleanup_thread.start()


def stop_handler_threads(app: flask.Flask) -> None:
    """
    Stop handler threads for background tasks.

    This function signals to the handle threads that they should stop
    processing tasks, then waits for them to finish before returning. If one
    of the handler threads blocks, this function will block as well.

    If background tasks are disabled, this function returns immediately.
    """
    if not app.config['ENABLE_BACKGROUND_TASKS']:
        return

    global should_stop
    should_stop = True

    # notify handler threads about having to stop
    wake_event.set()

    # then join handler threads
    # this is tried repeatedly, so that even if one thread is blocking, all others will be joined correctly
    running_threads = set(handler_threads)
    while running_threads:
        for handler_thread in running_threads.copy():
            handler_thread.join(1)
            if not handler_thread.is_alive():
                running_threads.remove(handler_thread)


def get_background_task_result(task_id: int, delete_when_final: bool = True) -> typing.Optional[dict[str, typing.Any]]:
    """
    Returns the result of a task.

    Creates a short evaluation of the specified task by returning the
    status and result as a dict. if specified, it also deletes
    the background task from the database.

    :param task_id: the id of the task
    :param delete_when_final: if true, the task will be deleted if it is in an final state (done/failed)
    :return: returns the status and the result of an background task as a dictionary
    """
    task: typing.Optional[BackgroundTask] = BackgroundTask.query.filter_by(id=task_id).first()
    if not task:
        return None

    result = {"status": task.status, "result": ""}

    if task.status.is_final():
        if task.type == 'dataverse_export':
            if task.result is None:
                result["result"] = ""
            elif task.status == BackgroundTaskStatus.DONE:
                result["result"] = task.result.get("url", "")
            else:
                result["result"] = task.result.get("error_message", "")

        if delete_when_final:
            db.session.delete(task)
            db.session.commit()

    return result


def _handle_background_tasks(app: flask.Flask, should_delete_expired_tasks: bool) -> None:
    with app.app_context():
        while not should_stop:
            if should_delete_expired_tasks:
                BackgroundTask.delete_expired_tasks()
            try:
                task = BackgroundTask.query.filter_by(status=BackgroundTaskStatus.POSTED).first()
            except Exception:
                # database might be unavailable for the moment, so no task to work on
                task = None
            if task is not None:
                if _claim_background_task(task):
                    status, result = _handle_background_task(task.type, task.data, task.id)
                    _set_background_task_status_with_result(task, status, result)
                else:
                    # another thread might have claimed the task first
                    continue
            else:
                wake_event.clear()
                wake_event.wait(TASK_WAIT_TIMEOUT)


def _claim_background_task(
        task: BackgroundTask
) -> bool:
    try:
        stmt = (
            db.update(
                BackgroundTask
            ).where(
                BackgroundTask.id == task.id,
                BackgroundTask.status == BackgroundTaskStatus.POSTED
            ).values(
                status=BackgroundTaskStatus.CLAIMED
            )
        )
        with db.engine.begin() as connection:
            updated_rowcount = connection.execute(stmt).rowcount
        return bool(updated_rowcount == 1)
    except Exception:
        # database might be temporarily unavailable, assume task was not claimed
        return False


def _handle_background_task(
        type: str,
        data: typing.Dict[str, typing.Any],
        task_id: typing.Optional[int]
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[typing.Dict[str, typing.Any]]]:
    handler = HANDLERS.get(type)
    try:
        if handler is not None:
            success, result = handler(data, task_id)
            if success:
                return BackgroundTaskStatus.DONE, result
            else:
                return BackgroundTaskStatus.FAILED, result
    except Exception:
        print("Exception during handler for task:\n", traceback.format_exc(), file=sys.stderr)
    return BackgroundTaskStatus.FAILED, None


def _set_background_task_status_with_result(
        task: BackgroundTask,
        task_status: BackgroundTaskStatus,
        result: typing.Optional[typing.Dict[str, typing.Any]]
) -> None:
    try:
        task.status = task_status
        if result is not None:
            if 'expiration_date' in result:
                task.expiration_date = result['expiration_date']
                del result['expiration_date']
            task.result = result
        if task_status.is_final() and task.auto_delete:
            db.session.delete(task)
        else:
            db.session.add(task)
        db.session.commit()
    except Exception:
        # task status could not be updated, no way to recover?
        pass


def reset_claimed_background_tasks() -> None:
    """
    Resets all currently claimed tasks to being posted.

    This should only be used before the handler threads are started, to allow
    retrying tasks which had been claimed when the app last ran, but did not
    finish before it was stopped.
    """
    claimed_tasks = BackgroundTask.query.filter_by(status=BackgroundTaskStatus.CLAIMED).all()
    for task in claimed_tasks:
        print(ansi_color(f"Resetting background task {task} to POSTED.", color=33), file=sys.stderr)
        task.status = BackgroundTaskStatus.POSTED
        db.session.add(task)
    db.session.commit()




import smtplib
import typing

import flask
import flask_mail

from ... import mail
from . import core
from ...models import BackgroundTask, BackgroundTaskStatus


def post_send_mail_task(
        subject: str,
        recipients: typing.List[str],
        text: str,
        html: str,
        auto_delete: bool = True
) -> typing.Tuple[BackgroundTaskStatus, typing.Optional[BackgroundTask]]:
    return core.post_background_task(
        type='send_mail',
        data={
            'subject': subject,
            'recipients': recipients,
            'text': text,
            'html': html
        },
        auto_delete=auto_delete
    )


def handle_send_mail_task(
        data: typing.Dict[str, typing.Any],
        task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    try:
        mail.send(flask_mail.Message(
            subject=data['subject'],
            sender=flask.current_app.config['MAIL_SENDER'],
            reply_to=flask.current_app.config['MAIL_REPLY_TO'],
            recipients=data['recipients'],
            body=data['text'],
            html=data['html']
        ))
        return True, None
    except smtplib.SMTPRecipientsRefused:
        return False, None



# coding: utf-8
"""
RESTful API for SampleDB
"""
import typing

import flask

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic.instruments import get_instrument, get_instruments
from ...logic import errors, utils, instruments

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def instrument_to_json(instrument: instruments.Instrument) -> typing.Dict[str, typing.Any]:
    return {
        'instrument_id': instrument.id,
        'name': utils.get_translated_text(instrument.name, 'en'),
        'description': utils.get_translated_text(instrument.description, 'en'),
        'is_hidden': instrument.is_hidden,
        'instrument_scientists': [user.id for user in instrument.responsible_users],
        'location_id': instrument.location_id
    }


class Instrument(Resource):
    @multi_auth.login_required
    def get(self, instrument_id: int) -> ResponseData:
        if flask.current_app.config['DISABLE_INSTRUMENTS']:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        try:
            instrument = get_instrument(
                instrument_id=instrument_id
            )
        except errors.InstrumentDoesNotExistError:
            return {
                "message": f"instrument {instrument_id} does not exist"
            }, 404
        return instrument_to_json(instrument)


class Instruments(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        if flask.current_app.config['DISABLE_INSTRUMENTS']:
            return []
        instruments = get_instruments()
        return [
            instrument_to_json(instrument)
            for instrument in instruments
        ]



# coding: utf-8
"""
Scripts for administrating SampleDB

Usage: sampledb <script> [options]
"""

import sys
import typing

from .scripts import script_documentation, script_functions


def help_and_exit(return_code: int) -> None:
    available_scripts = sorted(script_functions.keys())
    print(
        __doc__,
        'Available scripts:',
        ' - ' + '\n - '.join(available_scripts),
        '',
        'To find out how more about a script, use:',
        '',
        'python -m sampledb help <script>',
        '',
        sep='\n'
    )
    sys.exit(return_code)


def main(argv: typing.List[str]) -> None:
    if len(argv) < 2:
        return help_and_exit(1)
    script = argv[1]
    arguments = argv[2:]
    if script in ('help', '--help', '-h'):
        if arguments and arguments[0] in script_documentation:
            print(script_documentation[arguments[0]])
            sys.exit(0)
        return help_and_exit(0)
    if any(argument in ('help', '--help', '-h') for argument in arguments):
        if script in script_documentation:
            print(script_documentation[script])
            sys.exit(0)
        return help_and_exit(0)
    if script not in script_functions:
        return help_and_exit(2)
    script_functions[script](arguments)
    return None


if __name__ == '__main__':
    main(sys.argv)



import json
import typing
from http import HTTPStatus

import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField, SubmitField, SelectField
from wtforms.validators import DataRequired

from . import frontend, utils
from ..logic import errors, info_pages, users, languages, markdown_to_html, markdown_images
from ..models import User
from ..utils import FlaskResponseT


class InfoPageAcknowledgementForm(FlaskForm):
    info_page_ids = StringField(validators=[DataRequired()])


class SetInfoPageDisabledForm(FlaskForm):
    disabled = BooleanField()
    submit_set_info_page_disabled = SubmitField(validators=[DataRequired()])


class DeleteInfoPageForm(FlaskForm):
    submit_delete_info_page = SubmitField(validators=[DataRequired()])


@frontend.route('/info_pages/acknowledge', methods=['POST'])
@flask_login.login_required
def acknowledge_info_pages() -> FlaskResponseT:
    utils.check_current_user_is_not_readonly()
    form = InfoPageAcknowledgementForm()
    if not form.validate_on_submit():
        return flask.abort(HTTPStatus.BAD_REQUEST)
    existing_info_page_ids = {
        info_page.id
        for info_page in info_pages.get_info_pages()
    }
    info_page_ids = set()
    for info_page_id_str in form.info_page_ids.data.split(','):
        try:
            info_page_id = int(info_page_id_str)
        except ValueError:
            return flask.abort(HTTPStatus.BAD_REQUEST)
        if info_page_id not in existing_info_page_ids:
            return flask.abort(HTTPStatus.BAD_REQUEST)
        info_page_ids.add(info_page_id)
    info_pages.acknowledge_info_pages(
        info_page_ids=info_page_ids,
        user_id=flask_login.current_user.id,
    )
    return '', HTTPStatus.NO_CONTENT


@frontend.route('/admin/info_pages/')
@flask_login.login_required
def admin_info_pages() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    return flask.render_template(
        "info_pages/info_pages.html",
        info_pages=info_pages.get_info_pages(),
        url_map=flask.current_app.url_map,
        set_info_page_disabled_form=SetInfoPageDisabledForm(),
        delete_info_page_form=DeleteInfoPageForm(),
    )


@frontend.route('/admin/info_pages/<int:info_page_id>', methods=['GET', 'POST'])
@flask_login.login_required
def admin_info_page(info_page_id: int) -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    if flask.request.args.get('mode') == 'edit':
        return show_info_page_form(info_page_id)
    try:
        info_page = info_pages.get_info_page(info_page_id=info_page_id)
    except errors.InfoPageDoesNotExistError:
        return flask.abort(404)
    delete_info_page_form = DeleteInfoPageForm()
    if delete_info_page_form.validate_on_submit():
        utils.check_current_user_is_not_readonly()
        info_pages.delete_info_page(info_page_id=info_page_id)
        flask.flash(_('The info page has been deleted successfully.'), 'success')
        return flask.redirect(flask.url_for('frontend.admin_info_pages'))
    set_info_page_disabled_form = SetInfoPageDisabledForm()
    if set_info_page_disabled_form.validate_on_submit():
        utils.check_current_user_is_not_readonly()
        should_be_disabled = set_info_page_disabled_form.disabled.data
        if info_page.disabled == should_be_disabled:
            if should_be_disabled:
                flask.flash(_('The info page has already been disabled.'), 'warning')
            else:
                flask.flash(_('The info page has already been enabled.'), 'warning')
        else:
            info_pages.set_info_page_disabled(info_page_id=info_page_id, disabled=should_be_disabled)
            if should_be_disabled:
                flask.flash(_('The info page has been disabled successfully.'), 'success')
            else:
                flask.flash(_('The info page has been enabled successfully.'), 'success')
        return flask.redirect(flask.url_for('frontend.admin_info_page', info_page_id=info_page_id))
    return flask.render_template(
        "info_pages/info_page.html",
        info_page=info_page,
        url_map=flask.current_app.url_map,
        users=users.get_users(exclude_hidden=False, order_by=User.id, exclude_fed=True, exclude_eln_import=True),
        acknowledgements=info_pages.get_acknowledgements_for_info_page(info_page_id=info_page_id),
        set_info_page_disabled_form=set_info_page_disabled_form,
        delete_info_page_form=delete_info_page_form,
    )


class InfoPageForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    endpoint = SelectField()
    do_not_show_existing_users = BooleanField()
    clear_acknowledgements = BooleanField()


@frontend.route('/admin/info_pages/new', methods=['GET', 'POST'])
@flask_login.login_required
def admin_new_info_page() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    return show_info_page_form(None)


def show_info_page_form(info_page_id: typing.Optional[int]) -> FlaskResponseT:
    utils.check_current_user_is_not_readonly()

    class InfoPageTranslation(typing.TypedDict):
        language_id: int
        language: languages.Language
        title: str
        content: str

    info_page_translations: typing.List[InfoPageTranslation] = []
    info_page_language_ids: typing.List[int] = []
    info_page_form = InfoPageForm()
    url_rules_by_endpoint = info_pages.get_url_rules_by_endpoint()
    info_page_form.endpoint.choices = [
        ('', 'All routes')
    ] + sorted([
        (endpoint, ', '.join(rules))
        for endpoint, rules in url_rules_by_endpoint.items()
    ], key=lambda choice: choice[1])

    english = languages.get_language(languages.Language.ENGLISH)

    if info_page_id is not None:
        try:
            info_page = info_pages.get_info_page(info_page_id)
        except errors.InfoPageDoesNotExistError:
            return flask.abort(404)
        if 'action_submit' not in flask.request.form:
            info_page_translations = []
            info_page_language_ids = []
            info_page_language_codes = set()
            for translated_text in [info_page.title, info_page.content]:
                for language_code in translated_text:
                    info_page_language_codes.add(language_code)
            for language_code in info_page_language_codes:
                try:
                    language = languages.get_language_by_lang_code(language_code)
                except errors. LanguageDoesNotExistError:
                    continue
                info_page_language_ids.append(language.id)
                translation: InfoPageTranslation = {
                    'language': language,
                    'language_id': language.id,
                    'title': info_page.title.get(language_code, ''),
                    'content': info_page.content.get(language_code, ''),
                }
                info_page_translations.append(translation)
            info_page_form.endpoint.data = info_page.endpoint

    if info_page_id is None:
        current_page_url = flask.url_for('.admin_new_info_page')
    else:
        current_page_url = flask.url_for('.admin_info_page', info_page_id=info_page_id, mode='edit')
    if info_page_form.validate_on_submit():
        try:
            translation_data = json.loads(info_page_form.translations.data)
        except Exception:
            flask.flash(_('Please fill out the form.'), 'error')
            return flask.redirect(current_page_url)
        translation_keys = {'language_id', 'title', 'content'}
        if not isinstance(translation_data, list):
            translation_data = ()
        translated_title = {}
        translated_content = {}
        for translation in translation_data:
            if not isinstance(translation, dict):
                flask.flash(_('Please fill out the form.'), 'error')
                return flask.redirect(current_page_url)

            if set(translation.keys()) != translation_keys:
                flask.flash(_('Please fill out the form.'), 'error')
                return flask.redirect(current_page_url)

            language_id = int(translation['language_id'])
            try:
                language = languages.get_language(language_id)
            except errors.LanguageDoesNotExistError:
                continue
            title = translation['title'].strip()
            content = translation['content'].strip()
            if not title or not content:
                flask.flash(_('The fields for title and content are required.'), 'error')
                return flask.redirect(current_page_url)
            if len(title) > 100:
                flask.flash(_('Please enter a shorter title (at most 100 characters).'), 'error')
                return flask.redirect(current_page_url)
            translated_title[language.lang_code] = title
            translated_content[language.lang_code] = content
        if 'en' not in translated_title or 'en' not in translated_content:
            flask.flash(_('The fields for title and content are required.'), 'error')
            return flask.redirect(current_page_url)
        if info_page_id is None:
            info_page_id = info_pages.create_info_page(
                title=translated_title,
                content=translated_content,
                endpoint=info_page_form.endpoint.data or None,
                disabled=False
            ).id
            if info_page_form.do_not_show_existing_users.data:
                info_pages.acknowledge_info_page_for_existing_users(info_page_id)
            flask.flash(_('The info page has been created successfully.'), 'success')
        else:
            info_pages.update_info_page(
                info_page_id=info_page_id,
                title=translated_title,
                content=translated_content,
                endpoint=info_page_form.endpoint.data or None
            )
            if info_page_form.clear_acknowledgements.data:
                info_pages.clear_acknowledgements_for_info_page(info_page_id)
            flask.flash(_('The info page has been updated successfully.'), 'success')
        for content in translated_content.values():
            content_as_html = markdown_to_html.markdown_to_safe_html(content, anchor_prefix=f"info-page-content-{info_page_id}")
            markdown_images.mark_referenced_markdown_images_as_permanent(content_as_html)
        return flask.redirect(flask.url_for('.admin_info_page', info_page_id=info_page_id))
    return flask.render_template(
        'info_pages/info_page_form.html',
        info_page_form=info_page_form,
        info_page_translations=info_page_translations,
        info_page_language_ids=info_page_language_ids,
        languages=languages.get_languages(),
        ENGLISH=english,
        submit_text=_('Create') if info_page_id is None else _('Save')
    )



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .users import User
from .utils import Model


@enum.unique
class NotificationType(enum.Enum):
    OTHER = 0
    ASSIGNED_AS_RESPONSIBLE_USER = 1
    INVITED_TO_GROUP = 2
    INVITED_TO_PROJECT = 3
    ANNOUNCEMENT = 4
    RECEIVED_OBJECT_PERMISSIONS_REQUEST = 5
    INSTRUMENT_LOG_ENTRY_CREATED = 6
    REFERENCED_BY_OBJECT_METADATA = 7
    INSTRUMENT_LOG_ENTRY_EDITED = 8
    RESPONSIBILITY_ASSIGNMENT_DECLINED = 9
    REMOTE_OBJECT_IMPORT_FAILED = 10
    REMOTE_OBJECT_IMPORT_NOTES = 11
    AUTOMATIC_USER_FEDERATION = 12
    # Document any additional values in docs/administrator_guide/configuration.rst


@enum.unique
class NotificationMode(enum.Enum):
    IGNORE = 0
    WEBAPP = 1
    EMAIL = 2


class Notification(Model):
    __tablename__ = 'notifications'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[NotificationType] = db.Column(db.Enum(NotificationType), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    was_read: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Notification"]]

    def __init__(
            self,
            type: NotificationType,
            user_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            user_id=user_id,
            data=data,
            was_read=False,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, data={self.data})>'


class NotificationModeForType(Model):
    __tablename__ = 'notification_mode_for_types'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[NotificationType] = db.Column(db.Enum(NotificationType), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    mode: Mapped[NotificationMode] = db.Column(db.Enum(NotificationMode), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["NotificationModeForType"]]

    __table_args__ = (
        db.UniqueConstraint('type', 'user_id', name='_notification_mode_for_types_uc'),
    )

    def __init__(
            self,
            type: NotificationType,
            user_id: int,
            mode: NotificationMode
    ) -> None:
        super().__init__(
            type=type,
            user_id=user_id,
            mode=mode
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(type={self.type}, user_id={self.user_id}, mode={self.mode})>'



# coding: utf-8
"""
Add missing NOT NULL constraints to columns in authentications table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('authentications', 'user_id') and not column_is_nullable('authentications', 'login') and not column_is_nullable('authentications', 'type'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE authentications
            ALTER COLUMN user_id SET NOT NULL,
            ALTER COLUMN login SET NOT NULL,
            ALTER COLUMN type SET NOT NULL
    """))
    return True



# coding: utf-8
"""

"""

import enum
import typing

from sqlalchemy.orm import relationship, Query, Mapped

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .components import Component
    from .action_translations import ActionTranslation, ActionTypeTranslation
    from .instruments import Instrument
    from .users import User
    from .topics import Topic

usable_in_action_types_table = db.Table(
    'usable_in_action_types', db.metadata,
    db.Column('owner_usable_in_action_types_id', db.Integer, primary_key=True),
    db.Column('owner_action_type', db.Integer, db.ForeignKey('action_types.id')),
    db.Column('usable_in_action_types', db.Integer, db.ForeignKey('action_types.id'))
)

topic_action_association_table = db.Table(
    'action_topics',
    db.metadata,
    db.Column('action_id', db.Integer, db.ForeignKey('actions.id')),
    db.Column('topic_id', db.Integer, db.ForeignKey('topics.id'))
)


# defined here to avoid circular import dependency (actions -> scicat_export -> objects -> actions)
@enum.unique
class SciCatExportType(enum.Enum):
    RAW_DATASET = 0
    DERIVED_DATASET = 1
    SAMPLE = 2


class ActionType(Model):
    __tablename__ = 'action_types'
    __table_args__ = (
        db.UniqueConstraint('fed_id', 'component_id', name='action_types_fed_id_component_id_key'),
    )

    # default action type IDs
    # offset to -100 to allow later addition of new default action types
    # see: migrations/action_type_create_default_action_types.py
    SAMPLE_CREATION = -100 + 1
    MEASUREMENT = -100 + 2
    SIMULATION = -100 + 3
    TEMPLATE = -100 + 4

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    admin_only: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    show_on_frontpage: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    show_in_navbar: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    enable_labels: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_files: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_locations: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_publications: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_comments: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_activity_log: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_related_objects: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True)
    enable_project_link: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    disable_create_objects: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    is_template: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    usable_in_action_types: Mapped[typing.List['ActionType']] = relationship(
        'ActionType',
        secondary=usable_in_action_types_table,
        primaryjoin=id == usable_in_action_types_table.c.owner_action_type,
        secondaryjoin=id == usable_in_action_types_table.c.usable_in_action_types
    )
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    scicat_export_type: Mapped[typing.Optional[SciCatExportType]] = db.Column(db.Enum(SciCatExportType), nullable=True)
    translations: Mapped[typing.List['ActionTypeTranslation']] = relationship('ActionTypeTranslation', lazy='selectin')
    order_index: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    enable_instrument_link: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    show_in_object_filters: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True, server_default=db.true())

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ActionType"]]

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id!r})>'

    @property
    def name(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.name
            for translation in self.translations
            if translation.name
        }

    @property
    def description(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.description
            for translation in self.translations
            if translation.description
        }

    @property
    def object_name(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.object_name
            for translation in self.translations
            if translation.object_name
        }

    @property
    def object_name_plural(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.object_name_plural
            for translation in self.translations
            if translation.object_name_plural
        }

    @property
    def view_text(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.view_text
            for translation in self.translations
            if translation.view_text
        }

    @property
    def perform_text(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.perform_text
            for translation in self.translations
            if translation.perform_text
        }


class Action(Model):
    __tablename__ = 'actions'
    __table_args__ = (
        db.CheckConstraint(
            '(fed_id IS NOT NULL AND component_id IS NOT NULL) OR (type_id IS NOT NULL AND schema IS NOT NULL)',
            name='actions_not_null_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='actions_fed_id_component_id_key')
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey("action_types.id"), nullable=True)
    type: Mapped[typing.Optional[ActionType]] = relationship(ActionType, lazy='selectin')
    instrument_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey("instruments.id"), nullable=True)
    instrument: Mapped[typing.Optional['Instrument']] = relationship("Instrument", backref="actions", lazy='selectin')
    schema: Mapped[typing.Optional[typing.Dict[str, typing.Any]]] = db.Column(db.JSON, nullable=True)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey("users.id"), nullable=True)
    user: Mapped[typing.Optional['User']] = relationship("User", backref="actions", lazy='selectin')
    description_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    is_hidden: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    short_description_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False)
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')
    translations: Mapped[typing.List['ActionTranslation']] = relationship('ActionTranslation', lazy='selectin')
    admin_only: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    disable_create_objects: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    objects_readable_by_all_users_by_default: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    topics: Mapped[typing.List['Topic']] = relationship('Topic', secondary=topic_action_association_table, back_populates='actions')
    use_instrument_topics: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Action"]]

    def __init__(
            self,
            action_type_id: typing.Optional[int],
            schema: typing.Optional[typing.Dict[str, typing.Any]],
            instrument_id: typing.Optional[int] = None,
            user_id: typing.Optional[int] = None,
            description_is_markdown: bool = False,
            is_hidden: bool = False,
            short_description_is_markdown: bool = False,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None,
            admin_only: bool = False,
            disable_create_objects: bool = False,
            objects_readable_by_all_users_by_default: bool = False,
            use_instrument_topics: bool = False
    ) -> None:
        super().__init__(
            type_id=action_type_id,
            instrument_id=instrument_id,
            schema=schema,
            user_id=user_id,
            description_is_markdown=description_is_markdown,
            is_hidden=is_hidden,
            short_description_is_markdown=short_description_is_markdown,
            fed_id=fed_id,
            component_id=component_id,
            admin_only=admin_only,
            disable_create_objects=disable_create_objects,
            objects_readable_by_all_users_by_default=objects_readable_by_all_users_by_default,
            use_instrument_topics=use_instrument_topics
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(Action, other):
            return bool(
                self.id == other.id and
                self.instrument_id == other.instrument_id and
                self.user_id == other.user_id and
                self.description_is_markdown == other.description_is_markdown and
                self.short_description_is_markdown == other.short_description_is_markdown and
                self.is_hidden == other.is_hidden and
                self.schema == other.schema and
                self.fed_id == other.fed_id and
                self.component_id == other.component_id
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id!r})>'

    @property
    def name(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.name
            for translation in self.translations
            if translation.name
        }

    @property
    def description(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.description
            for translation in self.translations
            if translation.description
        }

    @property
    def short_description(self) -> typing.Dict[str, str]:
        return {
            translation.language.lang_code: translation.short_description
            for translation in self.translations
            if translation.short_description
        }



import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column("fed_identities", "login"):
        return False

    db.session.execute(db.text("""
        ALTER TABLE fed_identities
        ADD login BOOLEAN NOT NULL
        DEFAULT FALSE;
    """))
    return True



# coding: utf-8
"""
Script for setting whether a user is an administrator or not.

Usage: sampledb set_administrator <user_id> <yes_or_no>
"""

import sys
import typing

from .. import create_app
from ..logic.users import get_user, set_user_administrator
from ..logic.errors import UserDoesNotExistError


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 2 or not all(arguments) or arguments[1] not in ('yes', 'no'):
        print(__doc__)
        sys.exit(1)
    user_id_str, yes_or_no = arguments
    try:
        user_id = int(user_id_str)
    except ValueError:
        print("Error: user_id must be an integer", file=sys.stderr)
        sys.exit(1)
    is_admin = yes_or_no == 'yes'
    app = create_app()
    with app.app_context():
        try:
            user = get_user(user_id)
        except UserDoesNotExistError:
            print("Error: No user with this ID exists", file=sys.stderr)
            sys.exit(1)
        if user.is_admin == is_admin:
            print("Success: no change necessary")
        else:
            set_user_administrator(user.id, is_admin)
            if is_admin:
                print("Success: the user is an administrator now")
            else:
                print("Success: the user is a regular user now")



# coding: utf-8
"""
Add type_id column to locations table and set default location type.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..locations import LocationType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('locations', 'type_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE locations
            ADD type_id INTEGER NULL,
            ADD FOREIGN KEY(type_id) REFERENCES location_types(id)
    """))

    db.session.execute(db.text("""
        UPDATE locations
        SET type_id = :type_id
        WHERE type_id IS NULL
    """), params={
        'type_id': LocationType.LOCATION
    })

    db.session.execute(db.text("""
        ALTER TABLE locations
        ALTER type_id SET NOT NULL
    """))
    return True



# coding: utf-8
"""
Fix federation-related unique constraint for files table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE files
            DROP CONSTRAINT IF EXISTS files_fed_id_component_id_key
    """))
    db.session.execute(db.text("""
        ALTER TABLE files
            ADD CONSTRAINT files_fed_id_component_id_key UNIQUE(fed_id, object_id, component_id)
    """))
    return True



# coding: utf-8
"""

"""
import functools
import typing

import flask

_T = typing.TypeVar('_T')


def cache_per_request(key: typing.Optional[typing.Callable[..., typing.Any]] = None) -> typing.Callable[..., typing.Any]:
    def decorator(function: typing.Callable[..., _T]) -> typing.Callable[..., _T]:
        @functools.wraps(function)
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            if not hasattr(flask.g, 'cache'):
                flask.g.cache = {}
            cached_func = flask.g.cache.get(function, None)
            if cached_func is None:
                flask.g.cache[function] = {}
                cached_func = flask.g.cache[function]

            if key:
                dict_key = key(*args, **kwargs)
            else:
                dict_key = make_key(*args, **kwargs)

            if dict_key in cached_func:
                return cached_func[dict_key]

            res = function(*args, **kwargs)
            cached_func[dict_key] = res
            return res
        return wrapper
    return decorator


def make_key(
        *args: typing.Any,
        **kwargs: typing.Any
) -> typing.Tuple[
    typing.Tuple[typing.Any, ...],
    typing.FrozenSet[
        typing.Tuple[str, typing.Any]
    ]
]:
    return args, frozenset(kwargs.items())



# coding: utf-8
"""
Add eln_import_id and eln_object_id columns to objects_previous table and update NOT NULL constraint.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_names = db.session.execute(db.text("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'objects_previous'
    """)).fetchall()
    if ('eln_import_id',) in column_names:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE objects_previous
        ADD eln_import_id INTEGER,
        ADD eln_object_id VARCHAR,
        ADD FOREIGN KEY(eln_import_id) REFERENCES eln_imports(id)
    """))
    db.session.execute(db.text("""
        ALTER TABLE objects_previous
        DROP CONSTRAINT objects_previous_not_null_check
    """))
    db.session.execute(db.text("""
        ALTER TABLE objects_previous
        ADD CONSTRAINT objects_previous_not_null_check
            CHECK (
                (
                    fed_object_id IS NOT NULL AND
                    fed_version_id IS NOT NULL AND
                    component_id IS NOT NULL
                ) OR (
                    eln_import_id IS NOT NULL AND
                    eln_object_id IS NOT NULL
                ) OR (
                    action_id IS NOT NULL AND
                    data IS NOT NULL AND
                    schema IS NOT NULL AND
                    user_id IS NOT NULL AND
                    utc_datetime IS NOT NULL
                )
            )
    """))
    return True



# coding: utf-8
"""
Add use_real_orcid column to fed_user_aliases table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('fed_user_aliases', 'use_real_orcid'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE fed_user_aliases
        ADD use_real_orcid BOOLEAN NOT NULL DEFAULT(FALSE)
    """))
    return True




import typing

import flask
import flask_login
from flask_wtf import FlaskForm
from wtforms.fields import SelectField, IntegerField
from wtforms.validators import InputRequired
from flask_babel import _


from . import frontend
from .. import logic
from .utils import check_current_user_is_not_readonly
from ..utils import FlaskResponseT
from ..logic.utils import get_translated_text


class DeleteCategoryForm(FlaskForm):
    category_id = IntegerField(validators=[InputRequired()])


@frontend.route('/group_categories/')
@flask_login.login_required
def group_categories() -> FlaskResponseT:
    group_categories = list(logic.group_categories.get_group_categories())
    group_categories.sort(key=lambda category: get_translated_text(category.name).lower())
    group_categories_by_id = {
        category.id: category
        for category in group_categories
    }
    group_category_names = logic.group_categories.get_full_group_category_names()
    group_category_tree = logic.group_categories.get_group_category_tree()
    delete_category_form = DeleteCategoryForm()
    return flask.render_template(
        'group_categories/group_categories.html',
        group_categories=group_categories,
        group_categories_by_id=group_categories_by_id,
        group_category_names=group_category_names,
        group_category_tree=group_category_tree,
        delete_category_form=delete_category_form,
    )


@frontend.route('/group_categories/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_group_category() -> FlaskResponseT:
    if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_GROUP_CATEGORIES'] and not flask_login.current_user.is_admin:
        return flask.abort(403)
    return _show_group_category_form(None)


@frontend.route('/group_categories/<int:category_id>', methods=['GET', 'POST'])
@flask_login.login_required
def group_category(category_id: int) -> FlaskResponseT:
    if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_GROUP_CATEGORIES'] and not flask_login.current_user.is_admin:
        return flask.abort(403)

    try:
        logic.group_categories.get_group_category(category_id)
    except logic.errors.GroupCategoryDoesNotExistError:
        return flask.abort(404)

    delete_category_form = DeleteCategoryForm()
    if delete_category_form.validate_on_submit() and delete_category_form.category_id.data == category_id:
        check_current_user_is_not_readonly()
        logic.group_categories.delete_group_category(category_id=category_id)
        flask.flash(_('The group category was deleted successfully.'), 'success')
        return flask.redirect(flask.url_for('.group_categories'))

    return _show_group_category_form(category_id)


class GroupCategoryForm(FlaskForm):
    parent_category_id = SelectField()


def _show_group_category_form(category_id: typing.Optional[int]) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    group_category_form = GroupCategoryForm()

    group_category_names = logic.group_categories.get_full_group_category_names()
    group_category_form.parent_category_id.choices = [('-1', '—')] + [
        (str(category.id), ' / '.join(get_translated_text(name) for name in group_category_names[category.id]))
        for category in logic.group_categories.get_group_categories()
        if category_id is None or (category.id != category_id and not logic.group_categories.category_has_ancestor(category.id, category_id))
    ]

    has_error = False
    languages = logic.languages.get_languages()
    location_language_ids = set()
    language_id_by_lang_code = {
        language.lang_code: language.id
        for language in languages
    }
    english_id = logic.languages.Language.ENGLISH
    translation_language_ids = {english_id}
    translated_texts = {
        text_name: {
            'en': ''
        }
        for text_name in ('name',)
    }

    if not group_category_form.is_submitted():
        if category_id is not None:
            try:
                group_category = logic.group_categories.get_group_category(category_id)
            except logic.errors.LocationTypeDoesNotExistError:
                return flask.abort(404)
            # set fields from existing group category
            group_category_form.parent_category_id.data = str(group_category.parent_category_id)
            # set translated texts from existing group category
            for text_name in translated_texts:
                existing_translations = getattr(group_category, text_name)
                if existing_translations is not None:
                    for lang_code, translated_text in existing_translations.items():
                        translated_texts[text_name][lang_code] = translated_text
                        if lang_code in language_id_by_lang_code:
                            translation_language_ids.add(language_id_by_lang_code[lang_code])
        else:
            # set default values
            group_category_form.parent_category_id.data = '-1'
    else:
        translation_language_id_strs = flask.request.form.getlist('translation-languages')
        translation_language_ids = {
            language_id
            for language_id in language_id_by_lang_code.values()
            if str(language_id) in translation_language_id_strs or language_id == english_id
        }
        # set translated texts from form
        for text_name in translated_texts:
            for language in languages:
                if language.id in translation_language_ids:
                    translated_text = flask.request.form.get(f'{text_name}_{language.id}', '')
                    translated_texts[text_name][language.lang_code] = translated_text
                    location_language_ids.add(language.id)
                    if len(translated_text) == 0 or len(translated_text) > 100:
                        has_error = True

    if not has_error and group_category_form.validate_on_submit():
        if group_category_form.parent_category_id.data == '-1':
            parent_category_id = None
        else:
            parent_category_id = int(group_category_form.parent_category_id.data)
        if category_id is None:
            logic.group_categories.create_group_category(
                parent_category_id=parent_category_id,
                **translated_texts
            )
            flask.flash(_('The group category was created successfully.'), 'success')
        else:
            try:
                logic.group_categories.update_group_category(
                    category_id=category_id,
                    parent_category_id=parent_category_id,
                    **translated_texts
                )
            except logic.errors.CyclicGroupCategoryError:
                flask.flash(_('Invalid parent group category.'), 'error')
            else:
                flask.flash(_('The group category was updated successfully.'), 'success')
        return flask.redirect(flask.url_for('.group_categories'))

    return flask.render_template(
        'group_categories/group_category_form.html',
        group_category_form=group_category_form,
        translation_language_ids=list(translation_language_ids),
        translated_texts=translated_texts,
        ENGLISH=logic.languages.get_language(logic.languages.Language.ENGLISH),
        languages=languages,
        submit_text=_('Create') if category_id is None else _('Save')
    )



# coding: utf-8
"""
Logic module for management of actions

Actions are used to represent all kinds of methods or processes that result
in the creation of a new sample or measurement. What kind of object is
created when performing an action is defined by the action's type. The
action's schema defines what information should or may be recorded in SampleDB
for a newly created object.

Actions can be related to an instrument, which serves both to group actions
together and to provide instrument responsible users with permissions for
samples or measurements created with their instrument.

Actions can also be user-defined, to allow advanced users to create actions
for instruments which would otherwise not be included.

As actions form the basis for objects, they cannot be deleted. However, an
action can be altered as long as the type and instrument stay the same.
"""

import copy
import dataclasses
import typing

from .. import db
from .. import models
from . import errors, instruments, users, schemas, components, topics, favorites
from .utils import cache, get_translated_text
from .action_types import check_action_type_exists, ActionType


@dataclasses.dataclass(frozen=True)
class Action:
    """
    This class provides an immutable wrapper around models.actions.Action.
    """
    id: int
    type_id: typing.Optional[int]
    type: typing.Optional['ActionType']
    instrument_id: typing.Optional[int]
    instrument: typing.Optional[instruments.Instrument]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    user_id: typing.Optional[int]
    user: typing.Optional[users.User]
    is_hidden: bool
    name: typing.Dict[str, str]
    description: typing.Dict[str, str]
    description_is_markdown: bool
    short_description: typing.Dict[str, str]
    short_description_is_markdown: bool
    fed_id: typing.Optional[int]
    component_id: typing.Optional[int]
    component: typing.Optional[components.Component]
    admin_only: bool
    disable_create_objects: bool
    objects_readable_by_all_users_by_default: bool
    topics: typing.List[topics.Topic]
    use_instrument_topics: bool

    @classmethod
    def from_database(cls, action: models.Action) -> 'Action':
        return Action(
            id=action.id,
            type_id=action.type_id,
            type=ActionType.from_database(action.type) if action.type is not None else None,
            instrument_id=action.instrument_id,
            instrument=instruments.Instrument.from_database(action.instrument) if action.instrument is not None else None,
            schema=copy.deepcopy(action.schema) if action.schema is not None else None,
            user_id=action.user_id,
            user=users.User.from_database(action.user) if action.user is not None else None,
            is_hidden=action.is_hidden,
            name=action.name,
            description=action.description,
            description_is_markdown=action.description_is_markdown,
            short_description=action.short_description,
            short_description_is_markdown=action.short_description_is_markdown,
            fed_id=action.fed_id,
            component_id=action.component_id,
            component=components.Component.from_database(action.component) if action.component is not None else None,
            admin_only=action.admin_only,
            disable_create_objects=action.disable_create_objects,
            objects_readable_by_all_users_by_default=action.objects_readable_by_all_users_by_default,
            topics=[topics.Topic.from_database(topic) for topic in (
                action.instrument.topics if action.instrument is not None and action.use_instrument_topics else action.topics
            )],
            use_instrument_topics=action.use_instrument_topics
        )

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(id={self.id!r})>"


@typing.overload
def create_action(
        *,
        action_type_id: int,
        schema: typing.Dict[str, typing.Any],
        instrument_id: typing.Optional[int] = None,
        user_id: typing.Optional[int] = None,
        description_is_markdown: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: None = None,
        component_id: None = None,
        admin_only: bool = False,
        disable_create_objects: bool = False,
        objects_readable_by_all_users_by_default: bool = False,
        strict_schema_validation: bool = True,
        use_instrument_topics: bool = False,
) -> Action:
    ...


@typing.overload
def create_action(
        *,
        action_type_id: typing.Optional[int],
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        instrument_id: typing.Optional[int] = None,
        user_id: typing.Optional[int] = None,
        description_is_markdown: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: int,
        component_id: int,
        admin_only: bool = False,
        disable_create_objects: bool = False,
        objects_readable_by_all_users_by_default: bool = False,
        strict_schema_validation: bool = True,
        use_instrument_topics: bool = False,
) -> Action:
    ...


def create_action(
        *,
        action_type_id: typing.Optional[int],
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        instrument_id: typing.Optional[int] = None,
        user_id: typing.Optional[int] = None,
        description_is_markdown: bool = False,
        is_hidden: bool = False,
        short_description_is_markdown: bool = False,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None,
        admin_only: bool = False,
        disable_create_objects: bool = False,
        objects_readable_by_all_users_by_default: bool = False,
        strict_schema_validation: bool = True,
        use_instrument_topics: bool = False,
) -> Action:
    """
    Creates a new action with the given type and schema. If
    instrument_id is not None, the action will belong to the instrument with
    this ID.

    :param action_type_id: the ID of an existing action type
    :param schema: the schema for objects created using this action
    :param instrument_id: None or the ID of an existing instrument
    :param user_id: None or the ID of an existing user
    :param description_is_markdown: whether the description contains Markdown
    :param is_hidden: None or whether the action should be hidden
    :param short_description_is_markdown: whether the short description
        contains Markdown
    :param fed_id: the ID of the related action at the exporting component
    :param component_id: the ID of the exporting component
    :param admin_only: whether only admins may use the action to create objects
    :param disable_create_objects: whether the action may not be used to create objects
    :param objects_readable_by_all_users_by_default: whether objects created
        with this action should be readable by all signed-in users by default
    :param strict_schema_validation: whether schema validation should use strict mode
    :param use_instrument_topics: whether the topics of the instrument should be used
    :return: the created action
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    :raise errors.SchemaValidationError: when the schema is invalid
    :raise errors.InstrumentDoesNotExistError: when instrument_id is not None
        and no instrument with the given instrument ID exists
    :raise errors.UserDoesNotExistError: when user_id is not None and no user
        with the given user ID exists
    """
    assert (component_id is None) == (fed_id is None)
    assert component_id is not None or (action_type_id is not None and schema is not None)

    if action_type_id is not None:
        # ensure the action type exists
        check_action_type_exists(action_type_id)

    if schema is not None:
        schemas.validate_schema(schema, strict=strict_schema_validation)
    if instrument_id is not None:
        # ensure that the instrument can be found
        instruments.check_instrument_exists(instrument_id)
    if user_id is not None:
        # ensure that the user can be found
        users.check_user_exists(user_id)

    if component_id is not None:
        # ensure that the component can be found
        components.check_component_exists(component_id)

    action = models.Action(
        action_type_id=action_type_id,
        description_is_markdown=description_is_markdown,
        is_hidden=is_hidden,
        schema=schema,
        instrument_id=instrument_id,
        user_id=user_id,
        short_description_is_markdown=short_description_is_markdown,
        fed_id=fed_id,
        component_id=component_id,
        admin_only=admin_only,
        disable_create_objects=disable_create_objects,
        objects_readable_by_all_users_by_default=objects_readable_by_all_users_by_default,
        use_instrument_topics=use_instrument_topics if instrument_id is not None else False,
    )
    db.session.add(action)
    db.session.commit()
    return Action.from_database(action)


def get_actions(
        *,
        action_type_id: typing.Optional[int] = None,
        instrument_id: typing.Optional[int] = None,
) -> typing.List[Action]:
    """
    Returns all actions, optionally only actions of a given type.

    :param action_type_id: the ID of an existing action type, or None
    :param instrument_id: the ID of an existing instrument, or None
    :return: the list of actions
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    :raise errors.InstrumentDoesNotExistError: when no instrument with the
        given instrument ID exists
    """
    query = models.Action.query
    if action_type_id is not None:
        query = query.filter_by(type_id=action_type_id)
    if instrument_id is not None:
        query = query.filter_by(instrument_id=instrument_id)
    actions = query.all()
    if not actions:
        if action_type_id is not None:
            # ensure the action type exists
            check_action_type_exists(action_type_id=action_type_id)
        if instrument_id is not None:
            # ensure the instrument exists
            instruments.check_instrument_exists(instrument_id=instrument_id)
    return [
        Action.from_database(action)
        for action in actions
    ]


@cache
def check_action_exists(
        action_id: int
) -> None:
    """
    Check whether an action with the given action ID exists.

    :param action_id: the ID of an existing action
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    """
    if not db.session.query(db.exists().where(models.Action.id == action_id)).scalar():
        raise errors.ActionDoesNotExistError()


@cache
def get_action_owner_id(
        action_id: int
) -> typing.Optional[int]:
    """
    Get the owner of an action with a given ID.

    :param action_id: the ID of an existing action
    :return: the ID of the action's owner, or None
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    """
    row_or_none = db.session.query(models.Action.user_id).filter(models.Action.id == action_id).first()
    if row_or_none is None:
        raise errors.ActionDoesNotExistError()
    return typing.cast(typing.Optional[int], row_or_none[0])


def get_mutable_action(
        action_id: int,
        component_id: typing.Optional[int] = None
) -> models.Action:
    """
    Get the mutable action instance to perform changes in the database on.

    :param action_id: the ID of an existing action
    :param component_id: the ID of an existing component, or None
    :return: the mutable action
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    """
    action: typing.Optional[models.Action]
    if component_id is None:
        action = models.Action.query.filter_by(id=action_id).first()
    else:
        # ensure that the component can be found
        components.check_component_exists(component_id)
        action = models.Action.query.filter_by(fed_id=action_id, component_id=component_id).first()
    if action is None:
        raise errors.ActionDoesNotExistError()
    return action


def get_action(
        action_id: int,
        component_id: typing.Optional[int] = None
) -> Action:
    """
    Return the action with the given action ID.

    :param action_id: the ID of an existing action
    :param component_id: the ID of an existing component, or None
    :return: the action
    :raise errors.ActionDoesNotExistError: when no action with the given
        action ID exists
    """
    return Action.from_database(get_mutable_action(action_id, component_id))


def update_action(
        *,
        action_id: int,
        schema: typing.Optional[typing.Dict[str, typing.Any]],
        description_is_markdown: bool = False,
        is_hidden: typing.Optional[bool] = None,
        short_description_is_markdown: bool = False,
        admin_only: typing.Optional[bool] = None,
        disable_create_objects: typing.Optional[bool] = None,
        objects_readable_by_all_users_by_default: typing.Optional[bool] = None,
        use_instrument_topics: typing.Optional[bool] = None
) -> None:
    """
    Updates the action with the given action ID, setting its schema.

    :param action_id: the ID of an existing action
    :param schema: the new schema for objects created using this action
    :param description_is_markdown: whether the description contains Markdown
    :param is_hidden: None or whether the action should be hidden
    :param short_description_is_markdown: whether the short description
        contains Markdown
    :param admin_only: None or whether only admins may use the action to create objects
    :param disable_create_objects: None or whether the action may not be used to create objects
    :param objects_readable_by_all_users_by_default: whether objects created
        with this action should be readable by all signed-in users by default,
        or None
    :param use_instrument_topics: whether the topics of the instrument should be used, or None
    :raise errors.SchemaValidationError: when the schema is invalid
    :raise errors.InstrumentDoesNotExistError: when instrument_id is not None
        and no instrument with the given instrument ID exists
    """
    if schema is not None:
        schemas.validate_schema(schema, invalid_template_action_ids=[action_id], strict=True)
    action = get_mutable_action(action_id)
    action.description_is_markdown = description_is_markdown
    action.schema = schema
    action.short_description_is_markdown = short_description_is_markdown
    if is_hidden is not None:
        action.is_hidden = is_hidden
    if admin_only is not None:
        action.admin_only = admin_only
    if disable_create_objects is not None:
        action.disable_create_objects = disable_create_objects
    if objects_readable_by_all_users_by_default is not None:
        action.objects_readable_by_all_users_by_default = objects_readable_by_all_users_by_default
    if use_instrument_topics is not None:
        if action.instrument_id is not None or use_instrument_topics is False:
            action.use_instrument_topics = use_instrument_topics
    db.session.add(action)
    db.session.commit()
    update_actions_using_template_action(action_id)


def update_actions_using_template_action(
        template_action_id: int
) -> None:
    """
    Update the schemas of all actions using the given template action.

    :param template_action_id: the ID of a template action
    """
    template_action_schema = get_action(template_action_id).schema
    if template_action_schema is None:
        return
    template_action_schema = schemas.templates.process_template_action_schema(template_action_schema)
    actions = get_actions()
    updated_template_action_ids = []
    for action in actions:
        if action.id == template_action_id:
            continue
        if not action.schema:
            continue
        current_schema = copy.deepcopy(action.schema)
        updated_schema = schemas.templates.update_schema_using_template_action(current_schema, template_action_id, template_action_schema)
        if action.schema != updated_schema:
            try:
                schemas.validate_schema(updated_schema, strict=True)
            except errors.ValidationError:
                continue
            mutable_action = get_mutable_action(action.id)
            mutable_action.schema = updated_schema
            db.session.add(mutable_action)
            if action.type is not None and action.type.is_template:
                updated_template_action_ids.append(action.id)
    db.session.commit()
    for other_template_action_id in updated_template_action_ids:
        update_actions_using_template_action(other_template_action_id)


def get_action_type_ids_for_action_ids(
        action_ids: typing.Optional[typing.Sequence[int]]
) -> typing.Dict[int, typing.Optional[int]]:
    query = models.actions.Action.query.with_entities(
        models.actions.Action.id,
        models.actions.Action.type_id
    )
    if action_ids is not None:
        query = query.filter(
            models.actions.Action.id.in_(action_ids)
        )
    action_ids_and_action_type_ids = query.all()
    return {
        action_id: action_type_id
        for action_id, action_type_id in action_ids_and_action_type_ids
    }


def get_actions_for_topic(
        topic_id: int
) -> typing.List[Action]:
    """
    Get the list of actions assigned to a given topic.

    :param topic_id: the ID of an existing topic
    :return: the list of actions
    :raise errors.TopicDoesNotExistError: if the topic does not exist
    """
    actions = models.actions.Action.query.outerjoin(
        models.instruments.topic_instrument_association_table,
        models.Action.instrument_id == models.instruments.topic_instrument_association_table.c.instrument_id
    ).filter(
        db.or_(
            db.and_(
                models.Action.use_instrument_topics == db.false(),
                models.Action.topics.any(models.topics.Topic.id == topic_id)
            ),
            db.and_(
                models.Action.use_instrument_topics == db.true(),
                models.instruments.topic_instrument_association_table.c.topic_id == topic_id
            )
        )
    ).all()
    if not actions:
        topics.check_topic_exists(topic_id)
    return [
        Action.from_database(action)
        for action in actions
    ]


def sort_actions_for_user(
        actions: typing.Sequence[Action],
        user_id: int,
        sort_by_favorite: bool = True
) -> typing.List[Action]:
    """
    Sort an action list for a user.

    The actions are sorted by:
    - favorite / not favorite (if enabled)
    - action origin (local actions first)
    - user name (actions without users first)
    - instrument name (independent actions first)
    - action name

    :param actions: a list of actions to sort
    :param user_id: the ID of the user to sort actions for
    :param sort_by_favorite: whether favorite actions should be first in the sorted list
    :return: the sorted list of actions
    :raise errors.UserDoesNotExistError: if no user with the given user ID
        exists
    """
    if sort_by_favorite:
        user_favorite_action_ids = set(favorites.get_user_favorite_action_ids(user_id))
    else:
        user_favorite_action_ids = set()
    return sorted(actions, key=lambda action: (
        0 if action.id in user_favorite_action_ids else 1,
        0 if action.fed_id is None else 1,
        action.user.name.lower() if action.user and action.user.name is not None else '',
        get_translated_text(action.instrument.name).lower() if action.instrument else '',
        get_translated_text(action.name).lower()
    ))
